#!/usr/bin/env python
import argparse
from collections import defaultdict
import os
import glob

import numpy as np
import pandas as pd

from masato.get_abundance import get_otu_count, _agg_along_axis, _taxa_qc
from masato.utils import write_table


def _read_isolate_metadata(isolate_metadata_dir: str) -> pd.DataFrame:
    isolate_metadata_list = []
    for path in glob.glob(
        os.path.join(isolate_metadata_dir, "Destination * - * *.csv")
    ):
        dest_plate_barcode = os.path.basename(path).split(" ")[1]
        df = pd.read_csv(path, skiprows=2)
        df.iloc[:, 1] = df.iloc[:, 1].map(lambda x: x.split("_")[0])
        df["dest_plate"] = dest_plate_barcode
        isolate_metadata_list.append(df)
    return pd.concat(isolate_metadata_list, ignore_index=True)


def read_isolate_metadata_rich(
    isolate_metadata_dir: str, plate_metadata_path: str | None = None
) -> pd.DataFrame:
    isolate_metadata = _read_isolate_metadata(isolate_metadata_dir)
    isolate_metadata.columns = ["picking_coord", "src_plate", "dest_well", "dest_plate"]
    if plate_metadata_path is not None:
        plate_metadata = pd.read_csv(plate_metadata_path)
        if not plate_metadata.barcode.is_unique:
            raise ValueError("Plate metadata barcode column is not unique")
        isolate_metadata = isolate_metadata.merge(
            plate_metadata[
                ["barcode", "medium_type", "sample_type", "plate_name", "sample_group"]
            ].rename({"barcode": "src_plate"}, axis=1),
            on="src_plate",
            how="left",
        )
    else:
        isolate_metadata["medium_type"] = "unknown"
        isolate_metadata["sample_type"] = "unknown"
        isolate_metadata["plate_name"] = "unknown"
        isolate_metadata["sample_group"] = "unknown"
    isolate_metadata["dest_well_barcode"] = (
        isolate_metadata.dest_plate + "_" + isolate_metadata.dest_well
    )
    isolate_metadata["sample"] = isolate_metadata["dest_well_barcode"]
    isolate_metadata[["src_x", "src_y"]] = (
        isolate_metadata["picking_coord"]
        .str.extract(r"\((\d+.\d+), (\d+.\d+)\)")
        .astype(float)
        .to_numpy()
    )
    isolate_metadata["spike_in_16s"] = "genus;Sporosarcina,family;Chloroplast"
    isolate_metadata["spike_in_its"] = None
    return isolate_metadata.sort_values(
        ["sample_group", "sample_type", "medium_type", "src_plate", "sample"]
    ).set_index("sample")



def get_top_cols(
    df: pd.DataFrame,
    top_k: int = 5,
) -> pd.DataFrame:
    ret = {}
    argmax_arr = np.argsort(df, axis=1)
    max_arr = np.sort(df, axis=1)
    for i in range(min(top_k, df.shape[1])):
        ret[f"top_{i + 1}"] = df.columns[argmax_arr[:, -1 - i]]
        ret[f"top_{i + 1}_val"] = max_arr[:, -1 - i]
    return pd.DataFrame(ret, index=df.index)


def _calc_purity(
    df_zotu_count: pd.DataFrame, df_taxonomy: pd.DataFrame, level: str = "otu"
) -> pd.DataFrame:
    """Get purity table for input count table at a given taxonomy level"""
    df_rab = _taxa_qc(
        _agg_along_axis(
            df_zotu_count.div(df_zotu_count.sum(axis=1), axis=0).fillna(0),
            df_taxonomy[level],
            axis=1,
        ),
        rel_ab_thres=0,
        keep_rare=True,
        keep_unknown=True,
    )
    df_purity = get_top_cols(df_rab)
    df_purity["total_counts"] = df_zotu_count.sum(axis=1)
    return df_purity


def get_isolate_purity(
    df_zotu_cb: pd.DataFrame | None = None,
    df_zotu_cf: pd.DataFrame | None = None,
    df_taxonomy_b: pd.DataFrame | None = None,
    df_taxonomy_f: pd.DataFrame | None = None,
    levels: list = ["otu"],
) -> dict[str, list]:
    """Get purity table for multiple input count tables and the combined count table at
    multiple taxonomy levels

    If bacteria zotu count table is given, get its purity.
    If fungi zotu count table is given, get its purity.
    If both are given, get the purity for bacteria and fungi individually and collectively.
    """
    ret_purity = defaultdict(list)
    for level in levels:
        both = 0
        for df_zotu, df_tax in zip(
            [df_zotu_cb, df_zotu_cf], [df_taxonomy_b, df_taxonomy_f]
        ):
            if df_zotu is not None:
                df_purity = _calc_purity(df_zotu, df_tax, level)
                both += 1
                ret_purity[level].append(df_purity)
            else:
                ret_purity[level].append(None)
        if both == 2:
            df_zotu_count = pd.concat([df_zotu_cb, df_zotu_cf], axis=1).fillna(0)
            df_taxonomy = pd.concat([df_taxonomy_b, df_taxonomy_f], axis=0)
            df_purity = _calc_purity(df_zotu_count, df_taxonomy, level)
            ret_purity[level].append(df_purity)
        else:
            ret_purity[level].append(None)
    return ret_purity


# def get_isolate_purity(
#     df_isolate_meta_tsv: str,
#     df_zotu_count_bacteria_tsv: str = None,
#     df_zotu_count_fungi_tsv: str = None,
#     df_taxonomy_bacteria_tsv: str = None,
#     df_taxonomy_fungi_tsv: str = None,
#     level: str = "otu",
# ):
#     """Get purity table for isolates.

#     If bacteria zotu count table is given, get its purity.
#     If fungi zotu count table is given, get its purity.
#     If both are given, get the purity for bacteria and fungi individually and collectively.

#     If taxonomy tables are given, attach the taxonomy information to the purity table.
#     """
#     both = 0
#     ret_count = []
#     ret_tax = []
#     ret_purity = []
#     if df_zotu_count_bacteria_tsv is not None:
#         df_zotu_cb, _, df_taxonomy_b = get_otu_count(
#             df_zotu_count_bacteria_tsv,
#             df_isolate_meta_tsv,
#             df_taxonomy_bacteria_tsv,
#             warning=False,
#         )
#         df_rab_b = _taxa_qc(
#             _agg_along_axis(
#                 df_zotu_cb.div(df_zotu_cb.sum(axis=1), axis=0).fillna(0),
#                 df_taxonomy_b[level],
#                 axis=1,
#             ),
#             rel_ab_thres=0,
#             keep_rare=True,
#             keep_unknown=False,
#         )
#         df_purity_bacteria = get_top_cols(df_rab_b)
#         df_purity_bacteria["total_counts"] = df_zotu_cb.sum(axis=1)
#         both += 1
#         ret_count.append(df_zotu_cb)
#         ret_tax.append(df_taxonomy_b)
#         ret_purity.append(df_purity_bacteria)
#     else:
#         ret_count.append(None)
#         ret_tax.append(None)
#         ret_purity.append(None)
#     if df_zotu_count_fungi_tsv is not None:
#         df_zotu_cf, _, df_taxonomy_f = get_otu_count(
#             df_zotu_count_fungi_tsv,
#             df_isolate_meta_tsv,
#             df_taxonomy_fungi_tsv,
#             warning=False,
#         )
#         df_rab_f = _taxa_qc(
#             _agg_along_axis(
#                 df_zotu_cf.div(df_zotu_cf.sum(axis=1), axis=0).fillna(0),
#                 df_taxonomy_f[level],
#                 axis=1,
#             ),
#             rel_ab_thres=0,
#             keep_rare=True,
#             keep_unknown=False,
#         )
#         df_purity_fungi = get_top_cols(df_rab_f)
#         df_purity_fungi["total_counts"] = df_zotu_cf.sum(axis=1)
#         both += 1
#         ret_count.append(df_zotu_cf)
#         ret_tax.append(df_taxonomy_f)
#         ret_purity.append(df_purity_fungi)
#     else:
#         ret_count.append(None)
#         ret_tax.append(None)
#         ret_purity.append(None)
#     if both == 2:
#         df_zotu_count = pd.concat([df_zotu_cb, df_zotu_cf], axis=1).fillna(0)
#         df_taxonomy = pd.concat([df_taxonomy_b, df_taxonomy_f], axis=0)
#         df_rab = _taxa_qc(
#             _agg_along_axis(
#                 df_zotu_count.div(df_zotu_count.sum(axis=1), axis=0).fillna(0),
#                 df_taxonomy[level],
#                 axis=1,
#             ),
#             rel_ab_thres=0,
#             keep_rare=True,
#             keep_unknown=False,
#         )
#         df_purity = get_top_cols(df_rab)
#         df_purity["total_counts"] = df_zotu_count.sum(axis=1)
#         ret_count.append(df_zotu_count)
#         ret_tax.append(df_taxonomy)
#         ret_purity.append(df_purity)
#     else:
#         ret_count.append(None)
#         ret_tax.append(None)
#         ret_purity.append(None)
#     [df.sort_index(inplace=True) for df in ret_purity if df is not None]
#     try:
#         ret_count = pd.concat([df for df in ret_count if df is not None], axis=1)
#         ret_tax = pd.concat([df for df in ret_tax if df is not None], axis=0)
#     except ValueError:
#         ret_count = ret_tax = None
#     # drop columns of ret_count and rows of ret_tax whose suffix is ZOTU_UNKNWON
#     ret_count = ret_count.loc[:, ~ret_count.columns.str.endswith("ZOTU_UNKNOWN")].copy()
#     ret_tax = ret_tax.loc[~ret_tax.index.str.endswith("ZOTU_UNKNOWN"), :].copy()
#     return ret_count, ret_tax, pd.concat(ret_purity, axis=1)


def combine_count(
    df_isolate_meta_tsv: str,
    df_zotu_count_bacteria_tsv: str = None,
    df_zotu_count_fungi_tsv: str = None,
    df_taxonomy_bacteria_tsv: str = None,
    df_taxonomy_fungi_tsv: str = None,
) -> tuple[list[pd.DataFrame], list[pd.DataFrame]]:
    """This function simply combines the zotu count tables from two different
    experiments. We use it to merge amplicon sequencing data using 16S and ITS primer.

    It preprocesses the data according to what `get_otu_count` does, and
    renames ZOTU names that overlap between the two experiments.
    """
    dfs_zotu_c = []
    dfs_taxonomy = []
    for df_zotu_count_tsv, df_taxonomy_tsv in zip(
        [df_zotu_count_bacteria_tsv, df_zotu_count_fungi_tsv],
        [df_taxonomy_bacteria_tsv, df_taxonomy_fungi_tsv],
    ):
        if df_zotu_count_tsv is not None:
            df_zotu_c, _, df_taxonomy = get_otu_count(
                df_zotu_count_tsv, df_isolate_meta_tsv, df_taxonomy_tsv, warning=False
            )
            dfs_zotu_c.append(df_zotu_c)
            dfs_taxonomy.append(df_taxonomy)
        else:
            dfs_zotu_c.append(None)
            dfs_taxonomy.append(None)

    # rename ZOTU names by prefixing with 16S- for bacteria and ITS- for fungi
    if all(df is not None for df in dfs_zotu_c):
        overlap = set(dfs_zotu_c[0].columns) & set(dfs_zotu_c[1].columns)
        rename_dict_16s = {x: "16S-" + x for x in overlap}
        rename_dict_its = {x: "ITS-" + x for x in overlap}
        dfs_zotu_c[0].rename(columns=rename_dict_16s, inplace=True)
        dfs_zotu_c[1].rename(columns=rename_dict_its, inplace=True)
        dfs_taxonomy[0].rename(index=rename_dict_16s, inplace=True)
        dfs_taxonomy[1].rename(index=rename_dict_its, inplace=True)

    return dfs_zotu_c, dfs_taxonomy


def quality_control(
    df_purity: list[pd.DataFrame] | pd.DataFrame,
    purity_cutoff: float,
    total_counts_cutoff: int,
) -> pd.DataFrame:
    """Get the ZOTU count table for pure isolates in an isolate sequencing experiment.

    If df_purity is a list, we assume the dataframes have the same index, and a row
        is considered pure sa long as it is pure in ANY dataframes.

    Importantly, this function returns a binary count matrix whose index is from the
        index of `df_purity` and columns are from `top_1` column of `df_purity`.
        This is rather suboptimal. I need to think of an alternative way.
    """
    if isinstance(df_purity, list):
        df_purity = pd.concat(df_purity[["top_1", "top_1_val", "total_counts"]], axis=1)
    else:
        df_purity = df_purity[["top_1", "top_1_val", "total_counts"]]
    df_purity = df_purity.iloc[
        pd.concat(
            [
                (df_purity.iloc[:, i + 1] >= purity_cutoff)
                & (df_purity.iloc[:, i + 2] >= total_counts_cutoff)
                for i in range(0, df_purity.shape[1], 3)
            ],
            axis=1,
        )
        .any(axis=1)
        .tolist()
    ]
    top_1 = df_purity[["top_1"]].to_numpy()[
        np.arange(df_purity.shape[0]),
        df_purity[["top_1_val"]].to_numpy().argmax(axis=1),
    ]
    top_1 = pd.DataFrame(top_1, index=df_purity.index, columns=["#OTU ID"])
    return (
        top_1.reset_index(names="well_barcode")
        .assign(dummy=1)
        .pivot_table(
            index="well_barcode",
            columns="#OTU ID",
            values="dummy",
            fill_value=0,
        )
    ).astype(int)


def main():
# if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Utility functions for processing isolate metadata"
    )
    subparser = parser.add_subparsers(dest="subcommand")

    parser_isolate_meta = subparser.add_parser(
        "get_metadata",
        description="Summarize isolate metadata from the log files of isolate picking software",
    )
    parser_isolate_meta.add_argument(
        "-i",
        "--isolate_info_dir",
        help="Input directory containing the log file from isolate picking software",
        required=True,
        type=str,
        metavar="ISOLATE_METADATA_DIR",
    )
    parser_isolate_meta.add_argument(
        "-pm",
        "--plate_metadata_csv",
        help="Plate metadata CSV file",
        required=True,
        type=str,
        metavar="PLATE_METADATA_CSV",
    )
    parser_isolate_meta.add_argument(
        "-o",
        "--output_tsv",
        help="Output TSV file",
        required=True,
        type=str,
        metavar="OUTPUT_TSV",
    )

    parser_combine = subparser.add_parser(
        "combine_count",
        description="Combine zotu count tables",
    )
    parser_combine.add_argument(
        "-m",
        "--isolate_meta",
        help="Isolate metadata TSV file",
        required=True,
        type=str,
        metavar="ISOLATE_METADATA_TSV",
    )
    parser_combine.add_argument(
        "-cb",
        "--zotu_count_bacteria",
        help="Zotu count bacteria TSV file",
        default=None,
        type=str,
        metavar="ZOTU_COUNT_BACTERIA_TSV",
    )
    parser_combine.add_argument(
        "-cf",
        "--zotu_count_fungi",
        help="Zotu count fungi TSV file",
        default=None,
        type=str,
        metavar="ZOTU_COUNT_FUNGI_TSV",
    )
    parser_combine.add_argument(
        "-tb",
        "--taxonomy_bacteria",
        help="Taxonomy bacteria TSV file",
        default=None,
        type=str,
        metavar="TAXONOMY_BACTERIA_TSV",
    )
    parser_combine.add_argument(
        "-tf",
        "--taxonomy_fungi",
        help="Taxonomy fungi TSV file",
        default=None,
        type=str,
        metavar="TAXONOMY_FUNGI_TSV",
    )
    parser_combine.add_argument(
        "-pl",
        "--purity_levels",
        help="Taxonomic levels to calculate purity",
        default=[],
        type=str,
        nargs="+",
    )
    parser_combine.add_argument(
        "-cp",
        "--purity_cutoff",
        help="Purity cutoff",
        default=0.8,
        type=float,
    )
    parser_combine.add_argument(
        "-ct",
        "--total_counts_cutoff",
        help="Total counts cutoff",
        default=10,
        type=int,
    )
    parser_combine.add_argument(
        "-o",
        "--output_dir",
        help="Output directory",
        required=True,
        type=str,
        metavar="OUTPUT_DIR",
    )

    args = parser.parse_args()

    if args.subcommand == "get_metadata":
        isolate_info_dir = args.isolate_info_dir
        plate_metadata_csv = args.plate_metadata_csv
        output_tsv = args.output_tsv
        read_isolate_metadata_rich(isolate_info_dir, plate_metadata_csv).to_csv(
            output_tsv, sep="\t"
        )
    elif args.subcommand == "combine_count":
        (df_count_b, df_count_f), (df_taxon_b, df_taxon_f) = combine_count(
            args.isolate_meta,
            args.zotu_count_bacteria,
            args.zotu_count_fungi,
            args.taxonomy_bacteria,
            args.taxonomy_fungi,
        )
        os.makedirs(args.output_dir, exist_ok=True)
        if args.purity_levels:
            for l, (
                df_purity_bacteria,
                df_purity_fungi,
                df_purity,
            ) in get_isolate_purity(
                df_count_b,
                df_count_f,
                df_taxon_b,
                df_taxon_f,
                levels=args.purity_levels,
            ).items():
                if df_purity_bacteria is not None:
                    write_table(
                        df_purity_bacteria,
                        os.path.join(args.output_dir, f"purity_{l}_bacteria_only.tsv"),
                    )
                    otu_count_pure_b = (
                        quality_control(
                            df_purity_bacteria,
                            args.purity_cutoff,
                            args.total_counts_cutoff,
                        )
                        .reindex(df_count_b.index, fill_value=0)
                        .reindex(df_count_b.columns, axis=1, fill_value=0)
                    )
                    write_table(
                        otu_count_pure_b.transpose(),
                        os.path.join(args.output_dir, f"count_{l}_bacteria_only.biom"),
                    )
                if df_purity_fungi is not None:
                    write_table(
                        df_purity_fungi,
                        os.path.join(args.output_dir, f"purity_{l}_fungi_only.tsv"),
                    )
                    otu_count_pure_f = (
                        quality_control(
                            df_purity_fungi,
                            args.purity_cutoff,
                            args.total_counts_cutoff,
                        )
                        .reindex(df_count_f.index, fill_value=0)
                        .reindex(df_count_f.columns, axis=1, fill_value=0)
                    )
                    write_table(
                        otu_count_pure_f.transpose(),
                        os.path.join(args.output_dir, f"count_{l}_fungi_only.biom"),
                    )
                if df_purity is not None:
                    write_table(
                        df_purity, os.path.join(args.output_dir, f"purity_{l}.tsv")
                    )
                    otu_count_pure = (
                        quality_control(
                            df_purity, args.purity_cutoff, args.total_counts_cutoff
                        )
                        .reindex(df_count_b.index, fill_value=0)
                        .reindex(
                            df_count_b.columns.tolist() + df_count_f.columns.tolist(),
                            axis=1,
                            fill_value=0,
                        )
                        .transpose()
                    )
                    write_table(
                        otu_count_pure, os.path.join(args.output_dir, f"count_{l}.biom")
                    )
                    write_table(
                        otu_count_pure.loc[df_count_b.columns],
                        os.path.join(args.output_dir, f"count_{l}_bacteria.biom"),
                    )
                    write_table(
                        otu_count_pure.loc[df_count_f.columns],
                        os.path.join(args.output_dir, f"count_{l}_fungi.biom"),
                    )
        else:
            # simply concat them without any further modification, write data in the
            # same format as vsearch output
            df_zotu_count = (
                pd.concat([df_count_b, df_count_f], axis=1).fillna(0).transpose()
            )
            df_zotu_count.index.name = "#OTU ID"
            df_taxonomy = pd.concat([df_taxon_b, df_taxon_f], axis=0)
            write_table(df_zotu_count, os.path.join(args.output_dir, "count.biom"))
            df_taxonomy.to_csv(os.path.join(args.output_dir, "taxonomy.tsv"), sep="\t")
    else:
        raise ValueError("Invalid subcommand")



#!/usr/bin/env python
import argparse
import glob
import io
import os
import subprocess
import shutil
import gzip
from typing import IO
from io import StringIO
import tempfile
from concurrent.futures import as_completed
import json
from collections import defaultdict

import pandas as pd
from loky import get_reusable_executor
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from tqdm.auto import tqdm

from masato.utils import read_table, write_table


def merge_pairs(
    fastq_dir: str, fastq_out: str, num_threads: int, backend: str = "vsearch"
) -> None:
    # merge paired end reads
    output_dir = os.path.dirname(fastq_out)
    os.makedirs(output_dir, exist_ok=True)
    if backend == "usearch":
        subprocess.run(
            "usearch11 -fastq_mergepairs "
            f"{fastq_dir}/*_R1.fastq "
            "-relabel @ "
            f"-fastqout {output_dir}/merged.fastq "
            "-fastq_maxdiffs 25 "
            "-fastq_minovlen 50 "
            f"-report {output_dir}/merge_report.txt "
            f"-tabbedout {output_dir}/merge.tsv",
            shell=True,
        )
    elif backend == "vsearch":
        # vsearch cannot merge multiple files or perform relabeling at once, so we first
        # relabel with seqtk and then merge samples, and finally merge pairs with
        # vsearch.
        # Besides, vsearch does not have tsv or report output
        os.makedirs(f"{output_dir}/temp", exist_ok=True)
        fs = []
        ps = []

        # `fastq_dir` could be a path regex so that we can easily do merging for a
        # subset of samples in a directory.
        if os.path.isdir(fastq_dir):
            files = sorted(glob.glob(f"{fastq_dir}/*.fastq"))
            if not files:
                raise ValueError(f"No fastq files found in {fastq_dir}")
        else:
            files = [i for i in sorted(glob.glob(fastq_dir)) if i.endswith(".fastq")]
            if not files:
                raise ValueError(f"No fastq files found matching {fastq_dir}")
        if not all(f.endswith("_R1.fastq") or f.endswith("_R2.fastq") for f in files):
            raise ValueError(
                "Make sure to rename all fastq files of interest to end with _R1.fastq or _R2.fastq"
            )
        non_pair_files = []
        for f in files:
            if f.endswith("_R1.fastq"):
                if f"{f[:-9]}_R2.fastq" not in files:
                    non_pair_files.append(f)
            elif f.endswith("_R2.fastq"):
                if f"{f[:-9]}_R1.fastq" not in files:
                    non_pair_files.append(f)
        if non_pair_files:
            raise ValueError(
                f"Make sure that each sample has both _R1.fastq and _R2.fastq files. "
                f"The following files are not paired: {non_pair_files}"
            )

        batch = 1000
        for i in range(0, len(files), batch):
            file_list = files[i : i + batch]
            for f in file_list:
                f_base = os.path.basename(f)
                sample_name = f_base.split("_")[0]
                # use Popen for non-blocking execution
                outfile = open(f"{output_dir}/temp/{f_base}", "w")
                process = subprocess.Popen(
                    ["seqtk", "rename", f, f"sample={sample_name} "],
                    stdout=outfile,
                )
                fs.append(outfile)
                ps.append(process)
            [p.wait() for p in tqdm(ps)]
            [f.close() for f in fs]

        subprocess.run(
            f"cat {output_dir}/temp/*_R1.fastq > {output_dir}/temp_R1.fastq", shell=True
        )
        subprocess.run(
            f"cat {output_dir}/temp/*_R2.fastq > {output_dir}/temp_R2.fastq", shell=True
        )
        subprocess.run(
            [
                "vsearch",
                "--fastq_mergepairs",
                f"{output_dir}/temp_R1.fastq",
                "--reverse",
                f"{output_dir}/temp_R2.fastq",
                "--fastqout",
                fastq_out,
                # "--fastaout",
                # f"{output_dir}/merged.fa",
                "--fastq_maxdiffs",
                "25",
                "--fastq_minovlen",
                "50",
                "--threads",
                str(num_threads),
            ]
        )
        os.remove(f"{output_dir}/temp_R1.fastq")
        os.remove(f"{output_dir}/temp_R2.fastq")
        shutil.rmtree(f"{output_dir}/temp")


def qc(
    input_fastq: str,
    output_fasta: str | None = None,
    num_threads: int = 16,
    backend: str = "vsearch",
) -> str:
    """Output fasta is default to be modified from input fastq, but can be specified.
    For example, if input fastq file is /data/merged.fq.gz, then output fasta will be
    /data/merged.filtered.fq.gz, unless otherwise specified.
    """
    if output_fasta is None:
        output_dir = os.path.dirname(input_fastq)
        filename = os.path.basename(input_fastq).split(".")[0]
        output_fastx = os.path.join(output_dir, f"{filename}.filtered.fq")
    else:
        output_fastx = output_fasta.rstrip(".gz")

    if backend == "usearch":
        num_splits = None
        # Split the merged FASTQ file into multiple pieces
        total_lines = int(subprocess.check_output(["wc", "-l", input_fastq]).split()[0])
        lines_per_file = ((total_lines + num_splits - 1) // num_splits // 4) * 4
        base_dir = os.path.dirname(input_fastq)
        if num_splits > 1:
            subprocess.run(
                [
                    "split",
                    "-l",
                    str(lines_per_file),
                    "-d",
                    "--additional-suffix=.fastq",
                    input_fastq,
                    os.path.join(base_dir, "_split_"),
                ]
            )

            # Perform QC on each split FASTQ file
            with open(output_fastx, "w") as outfile:
                for split_file in glob.glob(os.path.join(base_dir, "_split_*.fastq")):
                    subprocess.run(
                        [
                            "usearch11",
                            "-fastq_filter",
                            split_file,
                            "-fastaout",
                            "/dev/stdout",
                            "-fastq_maxee",
                            "0.5",
                            "-fastq_minlen",
                            "200",
                            "-fastq_maxns",
                            "0",
                            "-relabel",
                            "filtered",
                            "-threads",
                            str(num_threads),
                        ],
                        stdout=outfile,
                    )
                    os.remove(split_file)
        else:
            subprocess.run(
                [
                    "usearch11",
                    "-fastq_filter",
                    input_fastq,
                    "-fastaout",
                    output_fastx,
                    "-fastq_maxee",
                    "0.5",
                    "-fastq_minlen",
                    "200",
                    "-fastq_maxns",
                    "0",
                    "-relabel",
                    "filtered",
                    "-threads",
                    str(num_threads),
                ]
            )
    elif backend == "vsearch":
        # we don't need to split anymore since there is no memory limit.
        subprocess.run(
            [
                "vsearch",
                "--fastq_filter",
                input_fastq,
                "--fastqout",
                output_fastx,
                "--fastq_maxee",
                "0.5",
                "--fastq_minlen",
                "100",
                "--fastq_maxns",
                "0",
                "--relabel",
                "filtered",
                # "--threads",
                # str(num_threads),
            ]
        )
        subprocess.run(["pigz", "-f", output_fastx])
        output_fastx += ".gz"
    return output_fastx


def split_fastq(args):
    # Calculate the number of lines in the FASTQ file (each read consists of 4 lines)
    total_lines = int(subprocess.check_output(["wc", "-l", args.input_file]).split()[0])
    # Calculate the number of lines for each split file, ensuring that the total number of lines is divisible by 4
    lines_per_file = ((total_lines + args.num_splits - 1) // args.num_splits // 4) * 4

    # Split the file using the `split` command
    subprocess.run(
        [
            "split",
            "-l",
            str(lines_per_file),
            "-d",
            "--additional-suffix=.fastq",
            args.input_file,
            os.path.join(args.output_dir, "split_"),
        ]
    )

    print(f"FASTQ file has been split into {args.num_splits} pieces.")


def add_depth_to_metadata(fastq_dir: str, metadata_path: str) -> None:
    # Generate metadata for the input FASTQ files

    # data = []
    # for f in os.listdir(fastq_dir):
    #     if f.endswith("_R1.fastq"):
    #         sample_name = f.replace("_R1.fastq", "")
    #         count = sum(1 for line in open(os.path.join(fastq_dir, f))) / 4
    #         data.append([sample_name, count])
    command = f"seqkit stats {fastq_dir}/*_R1.fastq -T"
    result = subprocess.run(command, capture_output=True, text=True, shell=True)

    # Check if the command was successful
    if result.returncode != 0:
        print("Error executing command:", result.stderr)
        exit()

    # Split the captured output into lines
    lines = result.stdout.splitlines()

    # Use pandas to read the tab-separated output into a DataFrame
    df_rc = pd.read_csv(io.StringIO("\n".join(lines)), sep="\t")
    df_rc["sample"] = df_rc.file.apply(
        lambda x: os.path.basename(x).replace("_R1.fastq", "")
    )
    df_rc = df_rc[["sample", "num_seqs"]].set_index("sample")
    df_rc.columns = ["read_count"]

    df = pd.read_table(metadata_path, index_col="sample")
    if "read_count" in df.columns:
        df.drop("read_count", axis=1, inplace=True)
    # raise error if there are samples in the metadata that are not in the fastq
    # directory, and raise warning if there are samples in the fastq directory that
    # are not in the metadata (these samples will be ignored)
    extra_samples = df_rc.index.difference(df.index)
    if len(extra_samples) > 0:
        print(
            f"WARNING: These samples in the fastq directory are not present in the metadata: {extra_samples}"
        )
    extra_samples = df.index.difference(df_rc.index)
    if len(extra_samples) > 0:
        raise ValueError(
            f"These samples in the metadata are not present in the fastq directory: {extra_samples}"
        )
    else:
        print("Great! All samples in the metadata are present in the fastq directory.")
    df_res = pd.merge(df, df_rc, left_index=True, right_index=True, how="inner")
    df_res.to_csv(metadata_path, sep="\t")


def subsample(
    fastq_dir: str,
    metadata_path: str,
    output_dir: str,
    output_metadata_path: str,
    num_subsamples: int,
    mode: str,
    seed: int = 100,
    metadata_only: bool = False,
) -> None:
    # Load the metadata CSV file into a pandas DataFrame
    df_meta = pd.read_table(metadata_path, index_col="sample")

    if mode == "reference":
        df_meta["read_count_subsample"] = df_meta["reference"].map(
            df_meta["read_count"]
        )
    elif mode == "min":
        df_meta["read_count_subsample"] = df_meta.read_count.min() - 1
    else:
        raise ValueError(f"Invalid mode: {mode}")

    os.makedirs(output_dir, exist_ok=True)

    # Begin subsampling
    subsample_metadata = []
    for sample_name, row in tqdm(df_meta.iterrows(), total=len(df_meta)):
        read_count = min(row["read_count_subsample"], row["read_count"])
        num_subs = (
            num_subsamples if row["read_count_subsample"] < row["read_count"] else 1
        )
        del row["read_count_subsample"], row["read_count"]
        ps = []
        for iteration in range(1, num_subs + 1):
            # Subsample R1 and R1 and output to the subsampled directory
            subsample_metadata.append(
                {
                    "sample": f"{sample_name}-subsampled-{iteration}",
                    "original_sample": sample_name,
                    "read_count": read_count,
                }
                | row.to_dict()
            )
            if metadata_only:
                continue
            # Increment the seed value
            seed += 1
            for suffix in ["_R1.fastq", "_R2.fastq"]:
                fastq_file = os.path.join(fastq_dir, f"{sample_name}{suffix}")
                if not os.path.isfile(fastq_file):
                    print(f"WARNING: {fastq_file} does not exist. Skipping.")
                    continue
                # can also use seqtk, but will need to redirect stdout to file
                p = subprocess.Popen(  # redirect stdout and stderr to /dev/null
                    [
                        "vsearch",
                        "--fastx_subsample",
                        fastq_file,
                        "--sample_size",
                        str(read_count),
                        "-randseed",
                        str(seed),
                        "--fastqout",
                        os.path.join(
                            output_dir, f"{sample_name}-subsampled-{iteration}{suffix}"
                        ),
                    ],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
                ps.append(p)
        [p.wait() for p in ps]
    pd.DataFrame(subsample_metadata).to_csv(output_metadata_path, sep="\t", index=False)


def db_construct(
    input_fastx: str, output_fasta: str, num_threads: int, backend: str = "vsearch"
) -> str:
    """Output fasta is default to be modified from input fastq, but can be specified.
    For example, if input fastq file is /data/merged.filtered.fq.gz, then output fasta
    will be /data/merged.uniq.fa.gz, unless otherwise specified.
    """
    if output_fasta is None:
        output_dir, base_name = os.path.split(input_fastx)
        base_name_noext = base_name.split(".")[0]
        output_path = f"{output_dir}/{base_name_noext}.uniq.fa"
    else:
        output_path = output_fasta.rstrip(".gz")

    if backend == "usearch":
        subprocess.run(
            [
                "usearch11",
                "-fastx_uniques",
                input_fastx,
                "-fastaout",
                f"{output_dir}/{base_name_noext}.uniq.fa",
                "-sizeout",
            ],
            check=True,
        )
        subprocess.run(
            [
                "usearch11",
                "-sortbysize",
                f"{output_dir}/{base_name_noext}.uniq.fa",
                "-fastaout",
                f"{output_dir}/{base_name_noext}.sorted.fa",
                "-minsize",
                "2",
            ],
            check=True,
        )
        subprocess.run(
            [
                "usearch11",
                "-search_exact",
                input_fastx,
                "-db",
                f"{output_dir}/{base_name_noext}.sorted.fa",
                "-strand",
                "both",
                "-notmatched",
                f"{output_dir}/{base_name_noext}.unmatched.fa",
                "-t",
                str(num_threads),
            ],
            check=True,
        )
    elif backend == "vsearch":
        subprocess.run(
            [
                "vsearch",
                "--fastx_uniques",
                input_fastx,
                "--fastaout",
                output_path,
                "--sizeout",
            ]
        )
        subprocess.run(["pigz", "-f", output_path])
        output_path += ".gz"
        return output_path


def cluster_uparse(
    input_fastq: str, db_fasta: str, output_dir: str, num_threads: int
) -> None:
    if output_dir is None:
        output_dir = os.path.dirname(input_fastq)
    os.makedirs(output_dir, exist_ok=True)
    subprocess.run(
        [
            "usearch11",
            "-cluster_otus",
            db_fasta,
            "-otus",
            f"{output_dir}/uparse_otu.fa",
            "-relabel",
            "OTU",
        ],
        check=True,
    )
    subprocess.run(
        [
            "usearch10",
            "-usearch_global",
            input_fastq,
            "-db",
            f"{output_dir}/uparse_otu.fa",
            "-strand",
            "both",
            "-id",
            "0.97",
            "-otutabout",
            f"{output_dir}/uparse_otu.tsv",
            "-biomout",
            f"{output_dir}/uparse_otu.biom",
            "-threads",
            str(num_threads),
            "-sample_delim",
            ".",
        ],
        check=True,
    )
    subprocess.run(
        [
            "usearch11",
            "-calc_distmx",
            f"{output_dir}/uparse_otu.fa",
            "-tabbedout",
            f"{output_dir}/uparse_otu.dist",
            "-maxdist",
            "0.2",
            "-termdist",
            "0.3",
            "-threads",
            str(num_threads),
        ],
        check=True,
    )
    subprocess.run(
        [
            "usearch11",
            "-cluster_aggd",
            f"{output_dir}/uparse_otu.dist",
            "-treeout",
            f"{output_dir}/uparse_otu.tree",
            "-clusterout",
            f"{output_dir}/uparse_otu.clust",
            "-id",
            "0.8",
            "-linkage",
            "min",
        ],
        check=True,
    )


def unoise3(
    input_fastq: str | IO[str] | gzip.GzipFile,
    output_fasta: str | IO[str] | gzip.GzipFile | None,
    min_size: int,
    min_length: int,
    relabel_prefix: str | None = None,
    size_out: bool = False,
    num_threads: int = 16,
    stderr: bool = True,
) -> None | str:
    stderr = subprocess.DEVNULL if not stderr else None
    input_arg, stdin = _decide_io_arg(input_fastq)
    output_arg, stdout = _decide_io_arg(output_fasta)

    args_qc = [
        "vsearch",
        "--fastq_filter",
        input_arg,
        "--fastqout",
        "-",
        "--fastq_maxee",
        "0.5",
        "--fastq_minlen",
        str(min_length),
        "--fastq_maxns",
        "0",
        "--relabel",
        "filtered",
        # "--threads",
        # str(num_threads),
    ]
    # print_command(args_qc)
    qc_proc = subprocess.Popen(
        args_qc,
        stdin=stdin,
        stdout=subprocess.PIPE,
        stderr=stderr,
    )

    args_uniq = [
        "vsearch",
        "--fastx_uniques",
        "-",
        "--fastaout",
        "-",
        "--sizeout",
    ]
    # print_command(args_uniq)
    uniq_proc = subprocess.Popen(
        args_uniq, stdin=qc_proc.stdout, stdout=subprocess.PIPE, stderr=stderr
    )

    args_unoise3 = [
        "vsearch",
        "--cluster_unoise",
        "-",
        "--minsize",
        str(min_size),
        "--sizeout",
        "--centroids",
        "-",
        "--threads",
        str(num_threads),
    ]
    # print_command(args_unoise3)
    unoise3_proc = subprocess.Popen(
        args_unoise3, stdin=uniq_proc.stdout, stdout=subprocess.PIPE, stderr=stderr
    )

    args_uchime3 = [
        "vsearch",
        "--uchime3_denovo",
        "-",
        "--nonchimeras",
        output_arg,
        "--relabel",
        relabel_prefix or "ZOTU",
    ]
    if size_out:
        args_uchime3.append("--sizeout")
    uchime3_proc = subprocess.Popen(
        args_uchime3, stdin=unoise3_proc.stdout, stdout=stdout, stderr=stderr
    )

    qc_proc.stdout.close()
    uniq_proc.stdout.close()
    unoise3_proc.stdout.close()
    if stdin is subprocess.PIPE:
        qc_proc.communicate(input_fastq.encode())
    uchime3_out, _ = uchime3_proc.communicate()
    if stdout is subprocess.PIPE:
        return uchime3_out.decode()


def search_global(
    input_fastq: str | gzip.GzipFile | IO[str] | None,
    zotu_fasta: str | gzip.GzipFile | IO[str] | None,
    output_tsv: str | gzip.GzipFile | IO[str] | None,
    not_matched_fasta: str | gzip.GzipFile | IO[str] | None = None,
    id_: float = 0.97,
    num_threads: int = 8,
    stderr: bool = True,
) -> None | str:
    """
    When `output_csv` is None, output is in captured in the stdout of returned process.
    When `not_matched_fasta` is None, unmatched query sequences are not saved as fasta.
    """
    fastq_arg, stdin = _decide_io_arg(input_fastq)
    db_arg, stdin_db = _decide_io_arg(zotu_fasta)
    tsv_arg, stdout = _decide_io_arg(output_tsv)
    if not_matched_fasta is None:
        not_matched_arg, stdout_not_matched = None, None
    else:
        not_matched_arg, stdout_not_matched = _decide_io_arg(not_matched_fasta)
    if stdin is not None and stdin_db is not None:
        raise ValueError(
            "`input_fastq` and `zotu_fasta` cannot both be file-like objects."
        )
    if stdout is not None and stdout_not_matched is not None:
        raise ValueError(
            "`output_tsv` and `not_matched_fasta` cannot both be file-like objects."
        )
    stderr = subprocess.DEVNULL if not stderr else None
    # https://github.com/torognes/vsearch/issues/552
    # https://github.com/torognes/vsearch/issues/467
    # https://github.com/torognes/vsearch/issues/392
    args_search = [
        "vsearch",
        "--usearch_global",
        fastq_arg,
        "--db",
        db_arg,
        "--strand",
        "plus",
        "--id",
        str(id_),
        "--maxaccepts",
        "10",
        "--maxhits",
        "1",
        "--threads",
        str(num_threads),
    ]
    if tsv_arg == "-":
        args_search.extend(["--otutabout", "-"])
    elif tsv_arg.endswith(".tsv") or tsv_arg.endswith(".tsv.gz"):
        args_search.extend(["--otutabout", tsv_arg])
    elif output_tsv.endswith(".biom"):
        args_search.extend(["--biomout", tsv_arg])
    else:
        raise ValueError(
            f"Invalid output file extension {output_tsv}. Must be one of .tsv, .biom"
        )
    if not_matched_arg is not None:
        args_search.extend(["--notmatched", not_matched_arg])
    sg_proc = subprocess.Popen(
        args_search,
        stdin=stdin,
        stdout=stdout,
        stderr=stderr,
    )
    sg_proc_out, sg_proc_err = sg_proc.communicate(
        input_fastq.encode() if stdin is subprocess.PIPE else None
    )
    if stdout is subprocess.PIPE:
        return sg_proc_out.decode()


def search_global_add_unknown(
    input_fastq: str,
    zotu_fasta: str,
    output_tsv: str,
    id_: float,
    unknown_name: str,
    num_threads: int,
) -> None:
    if not os.path.isfile(zotu_fasta):
        raise ValueError(f"ZOTU fasta file (database) {zotu_fasta} does not exist")
    if not os.path.isfile(input_fastq):
        raise ValueError(f"Query fastq file {input_fastq} does not exist")
    if output_tsv is None:
        output_dir = os.path.dirname(input_fastq)
        output_tsv = os.path.join(output_dir, "unoise3_zotu.biom")
    else:
        output_dir = os.path.dirname(output_tsv)
        os.makedirs(output_dir, exist_ok=True)
    output_not_matched = os.path.join(output_dir, "notmatched.fa")

    search_global(
        input_fastq, zotu_fasta, output_tsv, output_not_matched, id_, num_threads
    )

    # add a ZOTU_UNKNOWN to the zotu table by counting the number of reads that do not
    # matched to anything in database
    zotu_table = read_table(output_tsv)
    counter = {i: 0 for i in zotu_table.columns}
    for record in SeqIO.parse(output_not_matched, "fasta"):
        counter[record.id.split("=")[1]] += 1
    zotu_table.loc[unknown_name] = pd.Series(counter)
    write_table(zotu_table, output_tsv)


def _decide_io_arg(arg) -> tuple:
    if arg is None:  # PIPE as IO
        return "-", subprocess.PIPE
    elif isinstance(arg, str):
        if "\n" in arg:  # string literal corresponding to file content
            return "-", subprocess.PIPE
        else:  # file path
            arg = os.path.abspath(arg)
            if (
                not os.path.isfile(arg)
                and not os.path.isdir(arg)
                and not os.path.isdir(os.path.dirname(arg))
            ):
                raise ValueError(
                    f"{arg} does not appear to be stirng liternal and is also not a file path."
                )
            return arg, None
    elif isinstance(arg, StringIO):
        return "-", subprocess.PIPE
    else:
        if (
            not isinstance(arg, gzip.GzipFile)
            and not isinstance(arg, IO[str])
            and not isinstance(arg, subprocess.Popen)
        ):
            raise ValueError(f"Invalid argument {arg} of type {type(arg)}")
        return "-", arg


def _workflow_one_sample(
    seqs_sample: list[str],
    min_size: int,
    prefix: str | None = None,
    search: bool = True,
) -> tuple[list[str], list[str], list[int]]:
    num_qs = len(seqs_sample)
    input_fastq = "".join(seqs_sample)
    db_fasta = tempfile.NamedTemporaryFile()
    unoise3(
        input_fastq,
        db_fasta.name,
        min_size=min_size,
        relabel_prefix=prefix,
        size_out=False,
        num_threads=1,
        stderr=False,
    )
    try:
        zotus_name, zotus_seq = zip(
            *(
                (record.id, str(record.seq))
                for record in SeqIO.parse(db_fasta.name, "fasta")
            )
        )
    except ValueError:
        zotus_name, zotus_seq = [], []
    else:
        zotus_name, zotus_seq = list(zotus_name), list(zotus_seq)
    if zotus_name and search:
        count_tsv = search_global(
            input_fastq, db_fasta.name, None, num_threads=1, stderr=False
        )
        counts = pd.read_table(StringIO(count_tsv), index_col=0).iloc[:, 0].tolist()
        num_unknown = num_qs - sum(counts)
        if num_unknown > 0:
            zotus_name.append("#UNKNOWN")
            zotus_seq.append("#UNKNOWN")
            counts.append(num_unknown)
    else:
        counts = []
        if num_qs > 0:
            zotus_name = ["#UNKNOWN"]
            zotus_seq = ["#UNKNOWN"]
            counts = [num_qs]
    db_fasta.close()
    return zotus_name, zotus_seq, counts


def workflow_per_sample(
    input_fastq: str,
    output_path: str,
    min_size: int,
    prefix: str | None = None,
    num_threads: int = 8,
    search: bool = True,
) -> None:
    """Split input fastq file by sample and run a separate unoise3 workflow for each
        sample, so that each sample has one set of ZOTUs and counts.

    We assume sequences in the input fastq file are named like `@sample=<sample1>` and
        sequences from the same sample form consecutive blocks.

    If output path ends with json, save results in a json file like:
        `{<sample>: {"zotus": ["ATCG", "GCTA", ...], "counts": [10, 2, ...]}, ...}`.
    If output path ends with fasta, save results in a fasta file file like:
        ```
        >ATCG
        ATCG
        ```
        where sequence names are just sequences themselves. Identical sequences appear
        only once.
    """
    from masato.utils import smart_open

    current_sample = None
    fastq_for_current_sample = []
    future2sample = {}
    samples = set()
    # sample2future = {}
    results = {}

    executor = (
        get_reusable_executor(max_workers=num_threads, kill_workers=True, reuse=False)
        if num_threads > 1
        else None
    )
    f: IO[str] = smart_open(input_fastq, "r")
    for entry in zip(f, f, f, f):
        # Parse the header line to extract the sample name
        header: str = entry[0]
        sample_name = header.split()[0].split("=")[1].strip()
        if sample_name != current_sample and fastq_for_current_sample:
            if sample_name in samples:
                raise ValueError(
                    "Sequences in fastq file are not grouped by sample, found "
                    f"{sample_name} both before {current_sample} and after"
                )
            if executor:
                future = executor.submit(
                    _workflow_one_sample,
                    fastq_for_current_sample,
                    min_size,
                    prefix,
                    search,
                )
                future2sample[future] = current_sample
                # sample2future[current_sample] = future
            else:
                results[current_sample] = _workflow_one_sample(
                    fastq_for_current_sample, min_size, prefix, search
                )
            fastq_for_current_sample = []  # Reset for the next sample
            samples.add(current_sample)
        current_sample = sample_name
        fastq_for_current_sample.append("".join(entry))
    f.close()
    # Don't forget to process the last sample
    if fastq_for_current_sample:
        if executor:
            future = executor.submit(
                _workflow_one_sample, fastq_for_current_sample, min_size, prefix, search
            )
            future2sample[future] = current_sample
            # sample2future[current_sample] = future
            for future in tqdm(as_completed(future2sample), total=len(future2sample)):
                # start the task and save the future object
                sample = future2sample[future]
                results[sample] = future.result()
            results = {s: results[s] for s in samples}
        else:
            results[current_sample] = _workflow_one_sample(
                fastq_for_current_sample, min_size, prefix, search
            )
    # Assuming each task's result could be aggregated into a final result dictionary
    # This step depends on how you implement the process_queue and task results handling
    # You would collect results from each completed task here
    # This could involve collecting returned values or reading from a shared resource
    ext = os.path.splitext(output_path)[1]
    if ext == ".json":
        results = {s: {"zotus": z, "counts": c} for s, (_, z, c) in results.items()}
        with open(output_path, "w") as outfile:
            json.dump(results, outfile, indent=4)
    elif ext in set([".fa", ".fasta", ".fna"]):
        prefix = prefix or "ZOTU"
        zotus = {}
        for zns, zss, counts in results.values():
            for zn, zs, count in zip(zns, zss, counts):
                if zn == "#UNKNOWN":
                    continue
                zotus[zs] = zotus.get(zs, 0) + int(count.split("=")[1])
        # sort by frequency
        records = [
            SeqRecord(Seq(z), id=f"{prefix}{i}", description=str(zotus[z]))
            for i, z in enumerate(sorted(zotus, key=zotus.get, reverse=True), 1)
        ]
        with open(output_path, "w") as outfile:
            SeqIO.write(records, outfile, "fasta")
    else:
        raise ValueError(
            f"Invalid output file extension {ext}. Must be one of .json, .fa, .fasta, .fna"
        )


def aggregate_samples(
    input_json: str, output_fasta: str, output_count: str, prefix: str | None = None
) -> None:
    """Take the json file generated by workflow_per_sample and aggregate the results into
    fasta file of ZOTUs and count file.

    ZOTUs are labeled as <prefix><idx> where idx is determined by total counts across
    all samples, starting from <prefix>1 (e.g. 16S_ZOTU1).
    """
    if prefix is None:
        prefix = "ZOTU"
    # Load JSON data from file
    with open(input_json) as json_file:
        data = json.load(json_file)

    # Extract unique sequences and calculate their total counts, keeping track of "#UNKNOWN" separately
    sequence_counts = defaultdict(int)
    key_sequence_counts = defaultdict(lambda: defaultdict(int))
    unknown_counts = defaultdict(int)
    for key, value in data.items():
        for sequence, count in zip(value["zotus"], value["counts"]):
            if sequence == "#UNKNOWN":
                unknown_counts[key] = count
            else:
                sequence_counts[sequence] += count
                key_sequence_counts[key][sequence] = count

    # Sort sequences by total counts from high to low
    sorted_sequences = sorted(sequence_counts.items(), key=lambda x: x[1], reverse=True)

    # Create a list of SeqRecord objects with names ZOTU1, ZOTU2, ..., and total counts as description
    seq_records = []
    sequence_to_zotu = {}
    for i, (sequence, count) in enumerate(sorted_sequences, start=1):
        zotu_name = f"{prefix}{i}"
        seq_records.append(
            SeqRecord(Seq(sequence), id=zotu_name, description=f"total_counts={count}")
        )
        sequence_to_zotu[sequence] = zotu_name

    # Write to FASTA file
    with open(output_fasta, "w") as fasta_file:
        SeqIO.write(seq_records, fasta_file, "fasta")

    # Create a DataFrame with ZOTU names as index and counts as values
    df_data = defaultdict(list)
    index = []

    for sequence, zotu_name in sequence_to_zotu.items():
        index.append(zotu_name)
        for key in key_sequence_counts:
            df_data[key].append(key_sequence_counts[key].get(sequence, 0))

    # Add #UNKNOWN to the DataFrame as the last row
    # if "#UNKNOWN" in unknown_count:
    if unknown_counts:
        index.append(f"{prefix}_UNKNOWN")
        for key in key_sequence_counts:
            df_data[key].append(unknown_counts[key])
        # df_data[f"{prefix}_UNKNOWN"] = df_data.pop("#UNKNOWN")

    df = pd.DataFrame(df_data, index=index)
    df.index.name = "#OTU_ID"

    # Save the DataFrame to a CSV file (optional)
    write_table(df, output_count)


def cluster_unoise3(
    uniq_fasta: str, out_fasta: str, minsize: int, relabel: str = "ZOTU"
):
    if os.path.isdir(out_fasta):
        raise ValueError(f"Output fasta file {out_fasta} is a directory")
    output_dir = os.path.dirname(out_fasta)
    os.makedirs(output_dir, exist_ok=True)
    # zotu_init = os.path.join(output_dir, f"_{out_fasta}")
    # vsearch does not have `--otutab`, `--otutab_stats`, `--calc_distmx`, and
    # `--cluster_aggd`. As suggested on the github issue
    #  (https://github.com/torognes/vsearch/issues/392), we use `--search_global`
    # to replace `--otutab`, but the other commands are not implemented.
    # Besides, in usearch chimera is removed as part of the clustering step, but
    # in vsearch chimera removal is a separate step.
    unoise3_proc = subprocess.Popen(
        [
            "vsearch",
            "--cluster_unoise",
            uniq_fasta,
            "--minsize",
            str(minsize),
            "--centroids",
            # f"{output_dir}/temp.fa",
            "-",
            # "--strand",
            # "both",
        ],
        stdout=subprocess.PIPE,
    )
    uchime3_proc = subprocess.Popen(
        [
            "vsearch",
            "--uchime3_denovo",
            "-",
            "--nonchimeras",
            out_fasta,
            "--relabel",
            relabel,
        ],
        stdin=unoise3_proc.stdout,
    )
    uchime3_proc.wait()
    # os.remove(zotu_init)


def _fq2fa(input_fastq: str, output_fasta: str) -> None:
    """Convert fastq (could be gzipped) to output fasta (gzipped) using seqtk"""
    try:
        # Constructing the seqtk command
        cmd = f"seqtk seq -a {input_fastq} | pigz -f > {output_fasta}"

        # Executing the command
        subprocess.run(cmd, shell=True, check=True)
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Error executing command: {e.stderr}")


def _cluster_unoise3(
    input_fastq: str,
    db_fasta: str,
    minsize: int,
    id_: float,
    output_dir: str | None,
    num_threads: int,
    backend: str = "vsearch",
) -> None:
    if output_dir is None:
        output_dir = os.path.dirname(input_fastq)
    os.makedirs(output_dir, exist_ok=True)
    if backend == "usearch":
        subprocess.run(
            [
                "usearch11",
                "-unoise3",
                db_fasta,
                "-zotus",
                f"{output_dir}/unoise3_zotu.fa",
                "-tabbedout",
                f"{output_dir}/unoise3_zotu.tsv",
            ],
            check=True,
        )
        subprocess.run(
            [
                "usearch11",
                "-fastx_relabel",
                f"{output_dir}/unoise3_zotu.fa",
                "-prefix",
                "ZOTU",
                "-fastaout",
                f"{output_dir}/unoise3_zotu_relabel.fa",
                "-keep_annots",
            ],
            check=True,
        )
        os.rename(
            f"{output_dir}/unoise3_zotu_relabel.fa", f"{output_dir}/unoise3_zotu.fa"
        )
        subprocess.run(
            [
                "usearch10",
                "-otutab",
                input_fastq,
                "-db",
                f"{output_dir}/unoise3_zotu.fa",
                "-otutabout",
                f"{output_dir}/unoise3_zotu.tsv",
                "-biomout",
                f"{output_dir}/unoise3_zotu.biom",
                "-mapout",
                f"{output_dir}/unoise3_zotu.map",
                "-notmatched",
                f"{output_dir}/unoise3_zotu_notmatched.fa",
                "-dbmatched",
                f"{output_dir}/unoise3_zotu_dbmatched.fa",
                "-sizeout",
                "-strand",
                "both",
                "-threads",
                str(num_threads),
                "-sample_delim",
                ".",
            ],
            check=True,
        )
        subprocess.run(
            [
                "usearch10",
                "-otutab_stats",
                f"{output_dir}/unoise3_zotu.tsv",
                "-output",
                f"{output_dir}/unoise3_zotu_report.txt",
            ],
            check=True,
        )
        subprocess.run(
            [
                "usearch11",
                "-calc_distmx",
                f"{output_dir}/unoise3_zotu.fa",
                "-tabbedout",
                f"{output_dir}/unoise3_zotu.dist",
                "-maxdist",
                "0.2",
                "-termdist",
                "0.3",
            ],
            check=True,
        )
        subprocess.run(
            [
                "usearch11",
                "-cluster_aggd",
                f"{output_dir}/unoise3_zotu.dist",
                "-treeout",
                f"{output_dir}/unoise3_zotu.tree",
                "-clusterout",
                f"{output_dir}/unoise3_zotu.clust",
                "-id",
                "0.8",
                "-linkage",
                "min",
            ],
            check=True,
        )
    elif backend == "vsearch":
        # vsearch does not have `--otutab`, `--otutab_stats`, `--calc_distmx`, and
        # `--cluster_aggd`. As suggested on the github issue
        #  (https://github.com/torognes/vsearch/issues/392), we use `--search_global`
        # to replace `--otutab`, but the other commands are not implemented.
        # Besides, in usearch chimera is removed as part of the clustering step, but
        # in vsearch chimera removal is a separate step.
        subprocess.run(
            [
                "vsearch",
                "--cluster_unoise",
                db_fasta,
                "--minsize",
                str(minsize),
                "--centroids",
                f"{output_dir}/temp.fa",
                "--strand",
                "both",
            ],
            check=True,
        )
        subprocess.run(
            [
                "vsearch",
                "--uchime3_denovo",
                f"{output_dir}/temp.fa",
                "--nonchimeras",
                f"{output_dir}/unoise3_zotu.fa",
                "--relabel",
                "ZOTU",
            ],
        )
        os.remove(f"{output_dir}/temp.fa")
        # convert input fastq to fasta
        with open(f"{os.path.splitext(input_fastq)[0]}.fa", "w") as f:
            subprocess.run(["seqtk", "seq", "-A", input_fastq], stdout=f, check=True)
        subprocess.run(
            [
                "vsearch",
                "--usearch_global",
                f"{os.path.splitext(input_fastq)[0]}.fa",
                "--db",
                f"{output_dir}/unoise3_zotu.fa",
                "--strand",
                "plus",
                "--id",
                str(id_),
                "--otutabout",
                f"{output_dir}/unoise3_zotu.tsv",
                "--biomout",
                f"{output_dir}/unoise3_zotu.biom",
                "--alnout",
                f"{output_dir}/unoise3_zotu.aln",
                "--dbmatched",
                f"{output_dir}/unoise3_zotu_dbmatched.fa",
                "--dbnotmatched",
                f"{output_dir}/unoise3_zotu_dbnotmatched.fa",
                "--sizeout",
                "--threads",
                str(num_threads),
            ]
        )


def tax_nbc(
    input_fasta: str,
    db_fasta: str | None,
    output_path: str,
    num_threads: str,
    backend: str = "rdp_classifier",
) -> None:
    if backend == "usearch":
        if db_fasta is None:
            raise ValueError("db_fasta is required when using usearch")
        _ = subprocess.run(
            [
                "usearch11",
                "-nbc_tax",
                input_fasta,
                "-db",
                db_fasta,
                "-strand",
                "both",
                "-tabbedout",
                output_path,
                "-threads",
                str(num_threads),
            ],
            check=True,
        )
    elif backend == "rdp_classifier":
        if db_fasta is not None:
            print("WARNING: db_fasta is ignored when using rdp_classifier")
        subprocess.run(["classifier", "-o", output_path, input_fasta], check=True)
    else:
        raise ValueError(f"Invalid backend: {backend}")


def tax_sintax(
    input_fastq: str,
    db_udb: str,
    output_path: str,
    num_threads: str,
    backend: str = "vsearch",
) -> None:
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    if backend == "usearch":
        subprocess.run(
            [
                "usearch10",
                "-sintax",
                input_fastq,
                "-db",
                db_udb,
                "-strand",
                "both",
                "-sintax_cutoff",
                "0.8",
                "-threads",
                str(num_threads),
                "-tabbedout",
                output_path,
            ],
            check=True,
        )
    elif backend == "vsearch":
        subprocess.run(
            [
                "vsearch",
                "--sintax",
                input_fastq,
                "--db",
                db_udb,
                "--strand",
                "both",
                "--sintax_cutoff",
                "0.8",
                "--threads",
                str(num_threads),
                "--tabbedout",
                output_path,
            ],
            check=True,
        )


def main():
    parser = argparse.ArgumentParser(description="Perform QC on FASTQ files")
    subparsers = parser.add_subparsers(dest="subcommand")

    merge_pairs_parser = subparsers.add_parser(
        "merge_pairs", help="Merge paired-end reads"
    )
    merge_pairs_parser.add_argument(
        "-i", "--fastq_dir", help="Input directory containing paired-end fastq files"
    )
    merge_pairs_parser.add_argument(
        "-o", "--fastq_out", help="Output path for merged fastq file"
    )
    merge_pairs_parser.add_argument(
        "-t", "--num_threads", type=int, default=16, help="Number of threads to use"
    )

    db_construct_parser = subparsers.add_parser(
        "db_construct", help="Construct a database from a FASTA file"
    )
    db_construct_parser.add_argument(
        "-i", "--input_fastq", help="Input FASTA file to construct the database from"
    )
    db_construct_parser.add_argument(
        "-o",
        "--output_fasta",
        help="Output path for database FASTA file",
        type=str,
        default=None,
    )
    db_construct_parser.add_argument(
        "-t", "--num_threads", type=int, default=16, help="Number of threads to use"
    )

    cluster_uparse_parser = subparsers.add_parser(
        "cluster_uparse", help="Cluster reads using UPARSE"
    )
    cluster_uparse_parser.add_argument(
        "-i", "--input_fastq", help="Input FASTQ file to cluster", type=str
    )
    cluster_uparse_parser.add_argument(
        "-d", "--db_fasta", help="Input FASTA file to cluster against", type=str
    )
    cluster_uparse_parser.add_argument(
        "-o",
        "--output_dir",
        help="Output directory for UPARSE results",
        type=str,
        default=None,
    )
    cluster_uparse_parser.add_argument(
        "-t", "--num_threads", type=int, default=16, help="Number of threads to use"
    )

    cluster_unoise3_parser = subparsers.add_parser(
        "cluster_unoise3", help="Cluster reads using UNOISE3"
    )
    cluster_unoise3_parser.add_argument(
        "-i", "--uniq_fasta", help="Input FASTQ file to cluster", type=str
    )
    cluster_unoise3_parser.add_argument(
        "-m", "--minsize", type=int, default=8, help="Minimum cluster size"
    )
    cluster_unoise3_parser.add_argument(
        "-o", "--out_fasta", help="Output path for ZOTU fasta file"
    )
    unoise3_parser = subparsers.add_parser(
        "unoise3", help="Cluster reads using UNOISE3"
    )
    unoise3_parser.add_argument(
        "-i",
        "--input_fastq",
        help="Input FASTQ file to cluster",
        type=str,
        required=True,
    )
    unoise3_parser.add_argument(
        "-o",
        "--output_fasta",
        help="Output path for ZOTU fasta file",
        type=str,
        required=True,
    )
    unoise3_parser.add_argument(
        "-m", "--minsize", type=int, default=8, help="Minimum cluster size"
    )
    unoise3_parser.add_argument(
        "-l",
        "--min_length",
        type=int,
        default=100,
        help="Minimum length of sequences to cluster",
    )
    unoise3_parser.add_argument(
        "-l", "--relabel_prefix", type=str, default=None, help="Prefix for ZOTU labels"
    )
    unoise3_parser.add_argument(
        "-t", "--num_threads", type=int, default=4, help="Number of threads to use"
    )
    search_global_parser = subparsers.add_parser(
        "search_global", help="Search reads against a database"
    )
    search_global_parser.add_argument(
        "-i", "--input_fastq", help="Input FASTQ file to search", type=str
    )
    search_global_parser.add_argument(
        "-d", "--zotu_fasta", help="Input FASTA file to search against", type=str
    )
    search_global_parser.add_argument(
        "-o",
        "--output_tsv",
        help="Output path for ZOTU tsv file",
        type=str,
        default=None,
    )
    search_global_parser.add_argument(
        "--id", type=float, default=0.97, help="Minimum cluster identity"
    )
    search_global_parser.add_argument(
        "--unknown_name",
        type=str,
        default="ZOTU_UNKNOWN",
        help="Name for unknown ZOTUs",
    )
    search_global_parser.add_argument(
        "-t", "--num_threads", type=int, default=8, help="Number of threads to use"
    )
    workflow_per_sample_parser = subparsers.add_parser(
        "workflow_per_sample", help="Run UNOISE3 workflow per sample"
    )
    workflow_per_sample_parser.add_argument(
        "-i", "--input_fastq", help="Input FASTQ file to cluster", type=str
    )
    workflow_per_sample_parser.add_argument(
        "-o", "--output_path", help="Output path", type=str
    )
    workflow_per_sample_parser.add_argument(
        "-m", "--min_size", type=int, default=8, help="Minimum cluster size"
    )
    workflow_per_sample_parser.add_argument(
        "-l",
        "--relabel_prefix",
        type=str,
        default="ZOTU",
        help="Prefix for ZOTU labels",
    )
    workflow_per_sample_parser.add_argument(
        "-t", "--num_threads", type=int, default=16, help="Number of threads to use"
    )
    workflow_per_sample_parser.add_argument(
        "--search",
        action="store_true",
        help="Search reads against ZOTU database",
    )
    aggregate_samples_parser = subparsers.add_parser(
        "aggregate_samples", help="Aggregate ZOTU counts across samples"
    )
    aggregate_samples_parser.add_argument(
        "-i", "--input_json", help="Input JSON file with ZOTU counts", type=str
    )
    aggregate_samples_parser.add_argument(
        "-of", "--output_fasta", help="Output path for aggregated ZOTU fasta", type=str
    )
    aggregate_samples_parser.add_argument(
        "-oc", "--output_count", help="Output path for aggregated ZOTU counts", type=str
    )
    aggregate_samples_parser.add_argument(
        "-p",
        "--prefix",
        help="Prefix for ZOTU labels",
        type=str,
        default=None,
    )

    tax_nbc_parser = subparsers.add_parser(
        "tax_nbc", help="Taxonomy classification using NBC"
    )
    tax_nbc_parser.add_argument(
        "-i", "--input_fasta", help="Input FASTA file to classify", type=str
    )
    tax_nbc_parser.add_argument(
        "-d", "--db_fasta", help="Input FASTA file to classify against", type=str
    )
    tax_nbc_parser.add_argument(
        "-o", "--output_path", help="Output path for NBC results", type=str
    )
    tax_nbc_parser.add_argument(
        "-t", "--num_threads", type=int, default=16, help="Number of threads to use"
    )

    tax_sintax_parser = subparsers.add_parser(
        "tax_sintax", help="Taxonomy classification using SINTAX"
    )
    tax_sintax_parser.add_argument(
        "-i", "--input_fastq", help="Input FASTQ file to classify", type=str
    )
    tax_sintax_parser.add_argument(
        "-d", "--db_udb", help="Input udb file to classify against", type=str
    )
    tax_sintax_parser.add_argument(
        "-o", "--output_path", help="Output path for SINTAX results", type=str
    )
    tax_sintax_parser.add_argument(
        "-t", "--num_threads", type=int, default=16, help="Number of threads to use"
    )

    args = parser.parse_args()

    if args.subcommand == "db_construct":
        fastx_post_qc = qc(args.input_fastq, None, args.num_threads)
        db_construct(fastx_post_qc, args.output_fasta, args.num_threads)
    elif args.subcommand == "cluster_uparse":
        cluster_uparse(
            args.input_fastq, args.db_fasta, args.output_dir, args.num_threads
        )
    elif args.subcommand == "cluster_unoise3":
        cluster_unoise3(args.uniq_fasta, args.minsize, args.out_fasta)
    elif args.subcommand == "unoise3":
        unoise3(
            args.input_fastq,
            args.output_fasta,
            min_size=args.minsize,
            min_length=args.min_length,
            relabel_prefix=args.relabel_prefix,
            num_threads=args.num_threads,
        )
    elif args.subcommand == "search_global":
        search_global_add_unknown(
            args.input_fastq,
            args.zotu_fasta,
            args.output_tsv,
            id_=args.id,
            unknown_name=args.unknown_name,
            num_threads=args.num_threads,
        )
    elif args.subcommand == "workflow_per_sample":
        workflow_per_sample(
            args.input_fastq,
            args.output_path,
            min_size=args.min_size,
            prefix=args.relabel_prefix,
            num_threads=args.num_threads,
            search=args.search,
        )
    elif args.subcommand == "aggregate_samples":
        aggregate_samples(
            args.input_json, args.output_fasta, args.output_count, prefix=args.prefix
        )
    elif args.subcommand == "tax_nbc":
        tax_nbc(args.input_fasta, args.db_fasta, args.output_path, args.num_threads)
    elif args.subcommand == "tax_sintax":
        tax_sintax(args.input_fastq, args.db_udb, args.output_path, args.num_threads)
    else:
        raise ValueError(f"Invalid subcommand: {args.subcommand}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Simply run wget -qO- http://easy-amplicon-camii-test-data.s3.amazonaws.com/data.tar.gz | tar xz -C <user_specified_dir>
"""
import argparse
import subprocess


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--date", type=str)
    parser.add_argument("-o", "--output_dir", type=str)
    args = parser.parse_args()
    date = args.date
    output_dir = args.output_dir
    res = subprocess.run(
        f"wget -qO- http://easy-amplicon-camii-test-data.s3.amazonaws.com/data_{date}.tar.gz | tar xz -C {output_dir}",
        shell=True,
        check=True,
    )
    assert res.returncode == 0, f"Download failed with return code {res.returncode}"



import numpy as np
import pandas as pd
from scipy.cluster.hierarchy import linkage, leaves_list, fcluster
from sklearn.cluster import KMeans
from sklearn.datasets import make_blobs


def hierarchical_order(
    df: pd.DataFrame, num_clusters: int, method: str = "ward"
) -> (np.ndarray, np.ndarray):
    """
    Perform hierarchical clustering on the provided DataFrame and return the ordered indices and cluster labels.

    Parameters:
    df (pd.DataFrame): The DataFrame on which to perform hierarchical clustering.
    num_clusters (int): The number of clusters to divide the data into.
    method (str): The linkage method to use for clustering.

    Returns:
    tuple: A tuple containing:
           - np.ndarray: Array of ordered indices based on hierarchical clustering.
           - np.ndarray: Array of cluster labels for each row in the DataFrame.
    """
    linkage_result = linkage(df, method=method)
    ordered_indices = leaves_list(linkage_result)
    cluster_labels = (
        fcluster(linkage_result, num_clusters, criterion="maxclust")
        if num_clusters is not None
        else None
    )
    return ordered_indices, cluster_labels


def get_order(
    df: pd.DataFrame,
    num_clusters: int,
    inter_hier: bool = None,
    intra_hier: bool = None,
    seed: int = 42,
    method: str = "kmeans",
) -> (np.ndarray, np.ndarray):
    """
    Perform clustering and reorder rows of the dataframe based on hierarchical clustering or KMeans as specified.

    Parameters:
    df (pd.DataFrame): The dataframe to be clustered and reordered.
    num_clusters (int): The number of clusters to form.
    inter_hier (bool): Whether to reorder clusters based on hierarchical clustering of centroids for KMeans.
    intra_hier (bool): Whether to reorder within clusters based on hierarchical clustering for KMeans.
    seed (int): Random seed for reproducibility.
    method (str): Clustering method, 'kmeans' or 'hier' for hierarchical.

    Returns:
    tuple: A tuple containing:
           - np.ndarray: Array of row indices that reorder the dataframe rows.
           - np.ndarray: Array of cluster labels for each row in the input dataframe.
    """
    df = df.reset_index(drop=True)
    df_filled = df.fillna(0)

    if method == "kmeans":
        kmeans = KMeans(n_clusters=num_clusters, random_state=seed)
        cluster_labels = kmeans.fit_predict(df_filled)

        if inter_hier:
            centroids = kmeans.cluster_centers_
            ordered_indices, _ = hierarchical_order(pd.DataFrame(centroids), None)
            label_map = {
                old_label: new_label
                for new_label, old_label in enumerate(ordered_indices)
            }
            ordered_labels = np.array([label_map[label] for label in cluster_labels])
        else:
            ordered_labels = cluster_labels

        ordered_df = pd.DataFrame(df_filled).assign(label=ordered_labels)
        ordered_df.sort_values("label", inplace=True)
        new_indices = []
        if intra_hier:
            for label in np.unique(ordered_labels):
                cluster_indices = ordered_df[ordered_df["label"] == label].index
                cluster_data = df_filled.loc[cluster_indices]
                if len(cluster_data) == 1:
                    new_indices.extend(cluster_indices)
                else:
                    intra_cluster_order, _ = hierarchical_order(cluster_data, None)
                    new_indices.extend(cluster_indices[intra_cluster_order].tolist())
            new_indices = np.array(new_indices)
        else:
            new_indices = ordered_df.index.to_numpy()

    elif method == "hier":
        ordered_indices, cluster_labels = hierarchical_order(df_filled, num_clusters)
        ordered_df = df_filled.iloc[ordered_indices]
        new_indices = ordered_df.index.to_numpy()

    return np.array(new_indices), cluster_labels


if __name__ == "__main__":
    # Testing with simulated 2D Gaussian blobs
    X, y = make_blobs(n_samples=200, centers=4, cluster_std=0.60, random_state=0)
    blob_df = pd.DataFrame(X, columns=["Feature1", "Feature2"])

    # Apply the function
    indices, labels = get_order(
        blob_df, num_clusters=4, inter_hier=True, intra_hier=True, method="kmeans"
    )
    print(indices.shape, labels.shape)



#!/usr/bin/env python
"""
Simply run conda install --file conda_requirements.txt -c defaults -c bioconda -c conda-forge,
where the requirement file is in the root of the package. (This file in /src/masato/setup/conda_install.py)
"""

import argparse
import subprocess
import os


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--flavor", type=str, default="conda", choices=["conda", "mamba"]
    )
    args = parser.parse_args()
    flavor = args.flavor
    requirement_file = f"{os.path.dirname(__file__)}/../data/conda_requirements.txt"
    args = [
        flavor,
        "install",
        "--file",
        requirement_file,
        "-c",
        "defaults",
        "-c",
        "bioconda",
        "-c",
        "conda-forge",
        "-y",
    ]
    print(" ".join(args))
    ret = subprocess.run(args)
    assert (
        ret.returncode == 0
    ), f"Conda install failed with return code {ret.returncode}"



#!/usr/bin/env python3

import argparse

import pandas as pd


def parse_taxonomy_line(row):
    # Prepare the taxonomy and probability entries
    start_index = 5
    entries = iter(row[start_index:])
    # prob_entries = list(row[start_index::3])
    # tax_entries = list(row[start_index + 1 :: 3])
    # rank_entries = list(row[start_index + 2 :: 3])

    # Mapping for the taxonomic ranks
    tax_mapping = {
        "domain": "domain",
        "phylum": "phylum",
        "class": "class",
        "order": "order",
        "family": "family",
        "genus": "genus",
        "species": "species",
    }

    # Create an empty result dictionary
    result = {
        "otu": row[0],
        "domain": "unknown",
        "domain_p": -1,
        "phylum": "unknown",
        "phylum_p": -1,
        "class": "unknown",
        "class_p": -1,
        "order": "unknown",
        "order_p": -1,
        "family": "unknown",
        "family_p": -1,
        "genus": "unknown",
        "genus_p": -1,
        "species": "unknown",
        "species_p": -1,
    }

    # Populate the result dictionary with the provided data
    taxon_names = set()
    for tax, rank, prob in zip(entries, entries, entries):
        if rank in tax_mapping:
            if tax != "unknwon" and tax in taxon_names:
                tax += rank[0]
            tax = tax.replace(" ", "_")
            taxon_names.add(tax)
            result[tax_mapping[rank]] = tax
            result[tax_mapping[rank] + "_p"] = float(prob)

    return pd.Series(result)


def main():
# if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", type=str)
    parser.add_argument("-o", "--output", type=str)

    args = parser.parse_args()
    input_path = args.input
    output_path = args.output

    # Load the data from the tsv
    df = pd.read_table(input_path, header=None, names=list(range(30)))

    # Process each row
    df_res = df.apply(parse_taxonomy_line, axis=1)
    df_res.to_csv(output_path, sep="\t", index=False)



import pytest
from pathlib import Path


@pytest.fixture
def test_data_dir():
    return Path(__file__).parent / "data"


@pytest.mark.parametrize("amplicon_type", ["16s"])
@pytest.mark.parametrize("level", ["genus"])
@pytest.mark.parametrize(
    "fig_dir_order_pair",
    [("abc", "alphabetical"), ("clust", "hierarchical"), ("tax", "taxonomy_tree")],
)
@pytest.mark.parametrize("threshold_suffix_pair", [(1e-2, "1e-2"), (3, "top3")])
@pytest.mark.parametrize("subcommand", ["heatmap", "heatmap_log10", "stacked_bar"])
def test_plot(
    script_runner,
    test_data_dir,
    fig_dir_order_pair,
    subcommand,
    threshold_suffix_pair,
    amplicon_type,
    level,
):
    """
    Test script that mimics a complex nested loop from the original bash script.
    """

    # Unpack the parameters
    (fig_dir_name, order_option) = fig_dir_order_pair
    (abundance_threshold, suffix) = threshold_suffix_pair

    # Define test variables
    output_unoise3_dir = "output_unoise3_16s"

    # Prepare input and output directories
    input_file = test_data_dir / output_unoise3_dir / "unoise3_zotu.biom"
    metadata_file = test_data_dir / "metadata" / f"bulk_{amplicon_type}.tsv"
    taxonomy_file = (
        test_data_dir / output_unoise3_dir / "unoise3_zotu_rrndb_processed.tsv"
    )
    fig_dir = test_data_dir / "figs"
    adjusted_fig_dir = f"{fig_dir_name}_{suffix}"

    # Construct the command to run
    cmd = [
        "plot.py",
        "abundance_group",
        subcommand,
        "-i",
        str(input_file),
        "-m",
        str(metadata_file),
        "-t",
        str(taxonomy_file),
        "-f",
        str(fig_dir / f"abundance_{adjusted_fig_dir}_{amplicon_type}"),
        "-l",
        level,
        "-fo",
        order_option,
        "-a",
        str(abundance_threshold),
    ]

    # Run the script
    ret = script_runner.run(cmd)

    # Assertions
    assert ret.success, f"Script failed with return code {ret.returncode}"


@pytest.mark.parametrize("amplicon_type", ["16s"])
@pytest.mark.parametrize("level", ["genus"])
@pytest.mark.parametrize("rarefying", [False, True])  # To handle both commands
def test_plot_alpha_diversity(script_runner, test_data_dir, amplicon_type, level, rarefying):
    """
    Test script for plotting alpha diversity, with and without rarefying.
    """

    # Define test variables
    output_unoise3_dir = "output_unoise3_16s"
    fig_dir = test_data_dir / "figs"

    # Prepare input and output directories
    input_file = test_data_dir / output_unoise3_dir / "unoise3_zotu.biom"
    metadata_file = test_data_dir / "metadata" / f"bulk_{amplicon_type}.tsv"
    taxonomy_file = test_data_dir / output_unoise3_dir / "unoise3_zotu_rrndb_processed.tsv"

    # Construct the base command
    cmd = [
        "plot.py",
        "stats_sample_count",
        "-i", str(input_file),
        "-m", str(metadata_file),
        "-t", str(taxonomy_file),
        "-f", str(fig_dir / f"alpha_diversity_{'rarefying_' if rarefying else ''}{amplicon_type}"),
        "-l", "otu", level
    ]

    # Add rarefying options if needed
    if rarefying:
        cmd.extend(["-rv", "500", "-rn", "50"])

    # Run the script
    ret = script_runner.run(cmd)

    # Assertions
    assert ret.success, f"Script failed with return code {ret.returncode}"
    # output_file = fig_dir / f"alpha_diversity_{'rarefying_' if rarefying else ''}{amplicon_type}.png"
    # assert output_file.exists(), f"Output file not created: {output_file}"



#!/usr/bin/env python
"""
Simply run wget -qO- http://easy-amplicon-camii-test-data.s3.amazonaws.com/dist.tar.gz | tar xz -C masato.__path__[0] + "/../.."
"""

import subprocess

import masato


def main():
    output_dir = masato.__path__[0] + "/../.."

    res = subprocess.run(
        f"wget -qO- http://easy-amplicon-camii-test-data.s3.amazonaws.com/dist.tar.gz | tar xz -C {output_dir}", shell=True, check=True
    )
    assert res.returncode == 0, f"Download failed with return code {res.returncode}"

    # write to the info yaml file {output_dir}/rdp_classifier_2.14 to the rdp_classifier_path key
    # with open(INFO_PATH) as f:
    #     info = yaml.safe_load(f)
    # info["rdp_classifier_path"] = f"{output_dir}/rdp_classifier_2.14"
    # with open(INFO_PATH, "w") as f:
    #     yaml.dump(info, f)



"""
Example run:
python3 analysis_sintax.py \
    --input /home/ubuntu/data/carlotta_troubleshotting/sintax.tsv \
    --metadata /home/ubuntu/data/carlotta_troubleshotting/sample_metadata.tsv \
    --output_dir /home/ubuntu/data/carlotta_troubleshotting/res_sintax \
    --genus_spikein Sporosarcina \
    --group_key Group \
    --sample_count_thres 500 \
    --genus_rel_ab_thres 0.05 \
    --family_rel_ab_thres 0.05
"""

import argparse
from typing import Sequence

import numpy as np
import pandas as pd
import polars as pl
from skbio.diversity.alpha import shannon


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--input", type=str)
    parser.add_argument("--metadata", type=str)
    parser.add_argument("--output_dir", type=str)
    parser.add_argument("--genus_spikein", nargs="+", required=True, help="OTUs spikein.")
    parser.add_argument("--sample_weight_key", default=None)
    parser.add_argument("--group_key", default=None)
    parser.add_argument(
        "--sample_count_thres", type=int, required=True, help="Sample count threshold."
    )
    parser.add_argument(
        "--genus_rel_ab_thres",
        type=float,
        required=True,
        help="Genus relative abundance threshold.",
    )
    parser.add_argument(
        "--family_rel_ab_thres",
        type=float,
        required=True,
        help="Family relative abundance threshold.",
    )

    args = parser.parse_args()
    input_path = args.input
    metadata_path = args.metadata
    output_dir = args.output_dir
    genus_spikein = args.genus_spikein
    sample_weight_key = args.sample_weight_key
    group_key = args.group_key
    sample_count_thres = args.sample_count_thres
    genus_rel_ab_thres = args.genus_rel_ab_thres
    family_rel_ab_thres = args.family_rel_ab_thres

    sample_metadata = pd.read_table(metadata_path).rename(columns={"Sample": "sample"})
    if sample_weight_key is None:
        sample_weight_key = "sample_weight"
        sample_metadata[sample_weight_key] = 1

    sample_metadata = pl.from_pandas(sample_metadata).sort("sample")

    df = pl.read_csv(
        input_path,
        separator="\t",
        new_columns=["read_id", "prob", "strand", "assign"],
        has_header=False,
    )

    # numpy array is not a Sequence
    Sequence.register(np.ndarray)
    columns = np.char.array(
        ["domain", "phylum", "class", "order", "family", "genus", "species"]
    )

    df_level = df.select(
        pl.col("prob")
        .str.extract_all(r":([^()]+)")
        .list.eval(pl.element().str.lstrip(":"))
        .list.to_struct(fields=columns)
        .alias("level"),
    ).unnest("level")
    df_p = df.select(
        pl.col("prob")
        .str.extract_all(r"\(([\d.]+)\)")
        .list.eval(pl.element().str.slice(1, 4).cast(pl.Float64))
        .list.to_struct(fields=columns + "_p")
        .alias("p"),
    ).unnest("p")
    df_assign = df.select(
        pl.col("assign")
        .str.extract_all(r":([^,]+)")
        .list.eval(pl.element().str.lstrip(":"))
        .list.to_struct(fields=columns + "_conf"),
    ).unnest("assign")

    df_expand = pl.concat(
        [
            df.select(
                "read_id", pl.col("read_id").str.split(".").list.first().alias("sample")
            ),
            df_level,
            df_p,
            df_assign,
        ],
        how="horizontal",
    ).join(sample_metadata, on="sample")

    sample_count_melt = df_expand.groupby(["sample", "genus", "family"]).count()
    sample_qc = (
        sample_count_melt.groupby("sample")
        .agg(pl.sum("count"))
        .filter(pl.col("count") >= sample_count_thres)
        .select("sample")
        .to_series()
    )
    sample_count_melt = sample_count_melt.filter(pl.col("sample").is_in(sample_qc))
    sample_metadata = sample_metadata.filter(pl.col("sample").is_in(sample_qc))

    sample_count_spike_in = (
        sample_count_melt.filter(pl.col("genus").is_in(genus_spikein))
        .groupby("sample")
        .agg(pl.sum("count"))
        .sort("sample")
    )
    sample_count = sample_count_melt.filter(~pl.col("genus").is_in(genus_spikein))
    normalize_factor = sample_count_spike_in["count"] * sample_metadata[sample_weight_key]

    sample_genus_count = (
        sample_count.groupby(["sample", "genus"])
        .agg(pl.sum("count"))
        .pivot(values="count", index="sample", columns="genus")
        .sort("sample")
        .fill_null(0)
        .to_pandas()
        .set_index("sample")
    )
    sample_family_count = (
        sample_count.groupby(["sample", "family"])
        .agg(pl.sum("count"))
        .pivot(values="count", index="sample", columns="family")
        .sort("sample")
        .fill_null(0)
        .to_pandas()
        .set_index("sample")
    )

    sample_metadata = sample_metadata.to_pandas().set_index("sample")
    sample_metadata["spikein_reads"] = sample_count_spike_in["count"]
    sample_metadata["normalize_factor"] = normalize_factor

    # genus/family-level qc
    sample_genus_ab_rel = sample_genus_count.div(sample_genus_count.sum(axis=1), axis=0)
    sample_family_ab_rel = sample_family_count.div(sample_family_count.sum(axis=1), axis=0)
    genus_keep = sample_genus_ab_rel.sum(axis=0) >= genus_rel_ab_thres
    family_keep = sample_family_ab_rel.sum(axis=0) >= family_rel_ab_thres
    sample_genus_count = sample_genus_count.loc[:, genus_keep]
    sample_family_count = sample_family_count.loc[:, family_keep]

    # recalculate genus/family-level abundance
    sample_genus_ab_rel = sample_genus_count.div(sample_genus_count.sum(axis=1), axis=0)
    sample_family_ab_rel = sample_family_count.div(sample_family_count.sum(axis=1), axis=0)
    sample_genus_ab_abs = sample_genus_count.div(normalize_factor, axis=0)
    sample_family_ab_abs = sample_family_count.div(normalize_factor, axis=0)
    group_genus_ab_abs = (
        sample_genus_ab_abs.join(sample_metadata[group_key]).groupby(group_key).mean()
    )
    group_family_ab_abs = (
        sample_family_ab_abs.join(sample_metadata[group_key]).groupby(group_key).mean()
    )
    group_genus_ab_rel = (
        sample_genus_ab_rel.join(sample_metadata[group_key]).groupby(group_key).mean()
    )
    group_family_ab_rel = (
        sample_family_ab_rel.join(sample_metadata[group_key]).groupby(group_key).mean()
    )

    sample_genus_ab_abs.to_csv(f"{output_dir}/sample_genus_abs_ab.csv")
    sample_family_ab_abs.to_csv(f"{output_dir}/sample_family_abs_ab.csv")
    sample_genus_ab_rel.to_csv(f"{output_dir}/sample_genus_rel_ab.csv")
    sample_family_ab_rel.to_csv(f"{output_dir}/sample_family_rel_ab.csv")
    group_genus_ab_abs.to_csv(f"{output_dir}/group_genus_abs_ab.csv")
    group_family_ab_abs.to_csv(f"{output_dir}/group_family_abs_ab.csv")
    group_genus_ab_rel.to_csv(f"{output_dir}/group_genus_rel_ab.csv")
    group_family_ab_rel.to_csv(f"{output_dir}/group_family_rel_ab.csv")

    sample_metadata["shannon_genus"] = sample_genus_ab_rel.apply(shannon, axis=0)
    sample_metadata["shannon_family"] = sample_family_ab_rel.apply(shannon, axis=0)
    sample_metadata.to_csv(f"{output_dir}/sample_metadata.csv")



"""Parse the output of usearch -nbc_tax to extract taxonomic classification.

Two tsv files will be generated, one representing complete information in the input 
file and another only extract annotation at genus and family level. For the latter, 
any OTUs without assignment at genus or family level are discarded.
"""

import argparse
import re

import pandas as pd

print(
    "WARNING: This script is deprecated since it is for processing the output of "
    "`usearch -nbc_tax` command (which reimplements RPD classifier) and `usearch` is "
    "no longer used in this project. This script is here just for bookkeeping purpose. "
    "Use the original RDP classifier JAVA implementation instead."
)


def parse_taxonomy_line(row):
    # Define regex patterns
    prob_pattern = re.compile(r"([a-z]):([^,:]+)\((\d+\.\d+)\)")
    assign_pattern = re.compile(r"([a-z]):([^,:]+)")

    prob_matches = prob_pattern.findall(row["prob"])
    assign_matches = (
        assign_pattern.findall(row["assign"]) if pd.notna(row["assign"]) else []
    )

    # Create an empty result dictionary
    result = {
        "otu": row["otu"],
        "domain": None,
        "domain_p": None,
        "domain_conf": None,
        "phylum": None,
        "phylum_p": None,
        "phylum_conf": None,
        "class": None,
        "class_p": None,
        "class_conf": None,
        "order": None,
        "order_p": None,
        "order_conf": None,
        "family": None,
        "family_p": None,
        "family_conf": None,
        "genus": None,
        "genus_p": None,
        "genus_conf": None,
        "species": None,
        "species_p": None,
        "species_conf": None,
    }

    # Mapping for the taxonomic ranks
    tax_mapping = {
        "d": "domain",
        "p": "phylum",
        "c": "class",
        "o": "order",
        "f": "family",
        "g": "genus",
        "s": "species",
    }

    # Populate the result dictionary with classification matches
    for rank, tax, prob in prob_matches:
        if rank in tax_mapping:
            result[tax_mapping[rank]] = tax
            result[tax_mapping[rank] + "_p"] = float(prob)
    # Populate the result dictionary with confidence based on assign
    for rank, tax in assign_matches:
        if rank in tax_mapping and result[tax_mapping[rank]] == tax:
            result[tax_mapping[rank] + "_conf"] = tax

    return pd.Series(result)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input", type=str)
    parser.add_argument("output", type=str)
    parser.add_argument("output_simple", type=str)

    args = parser.parse_args()
    input_path = args.input
    output_path = args.output
    output_simple_path = args.output_simple

    # Load the data from the tsv
    df = pd.read_csv(
        input_path,
        sep="\t",
        names=["otu", "prob", "strand", "assign"],
    )

    # Process each row
    df_res = df.apply(parse_taxonomy_line, axis=1)
    df_res.to_csv(
        output_path,
        sep="\t",
        index=False,
    )

    df_res = df_res[["otu", "genus_conf", "family_conf"]].dropna()
    df_res.columns = ["otu", "genus", "family"]
    df_res.to_csv(
        output_simple_path,
        sep="\t",
        index=False,
    )



#!/usr/bin/env python3
"""
Take a OTU count table, sample metadata, and OTU taxonomy table as input, output 
tables of relative/absolute abundance at OTU level and relative abundance at 
genus/family level.

QC is also performed during the process. Poor samples with low total OTU counts are 
removed, and poor genus and family with low total relative abundance are removed.

Spike-in genus needs to be specified as an argument. Sample weight could also be 
provided as an argument. Spike-in genus won't be considered in the final tables. 

Absolute OTU abundance is calculated by dividing OTU counts by spike-in counts and 
sample weight.

Relative OTU abundance is calculated by normalizing absolute OTU abundance to 1.

Relative genus/family abundance is calculated by first averaging over any replicates 
within group and averaging again relative OTU abundance of OTUs corresponding to the 
same genus/family.

Note that rows of OTU abundance tables are sample, while rows of genus/family abundance 
tables are groups.

Example run:
get_abundance.py \
  --otu_count_tsv_path outputs_unoise3/unoise3_zotu.tsv \
  --metadata_path metadatas/all.tsv \
  --otu_taxonomy_path outputs_unoise3/unoise3_zotu_rrndb_processed.tsv \
  --output_dir outputs_rrndb/all \
  --spikein_taxa_key spike_in \
  --rep_group_key rep_group \
  --tax_levels otu genus family order \
  --rel_ab_thres 0.01 \
  --keep_others \
  --keep_unknown
"""

import os
import argparse

import pandas as pd
import polars as pl

from masato.utils import read_table, read_table


def _find_otus_by_taxon(df_tax: pd.DataFrame, taxon: str) -> list[str]:
    # index of df_tax are OTU numbers
    level, name = taxon.split(";")
    return df_tax.query(f"{level} == '{name}'").index.tolist()


def _calc_norm_factor(
    df_otu_count: pd.DataFrame,
    df_meta: pd.DataFrame,
    df_tax: pd.DataFrame,
    sample_weight_key: str = None,
    spikein_taxa_key: str = None,
) -> pd.DataFrame:
    """
    Get absolute and relative abundance of OTUs. In this step replicate samples are
        still not averaged.

    The index of `df_otu_count` should be sample names, and columns should be OTU
        numbers.
    The index of `df_tax` should be OTU numbers.
    The index of `df_meta` should be sample names. Only samples specified in `df_meta`
        will be kept in the output.

    A new metadata dataframe will be returned.
    """
    spikein2otus = find_spikein_otus(df_meta, df_tax, spikein_taxa_key)

    def compute_spikein_and_norm(row: pd.Series) -> pd.Series:
        otu_count_series = df_otu_count.loc[row.name]

        if spikein_taxa_key and isinstance(row[spikein_taxa_key], str):
            otus_spikein = [
                otu
                for spikein in row[spikein_taxa_key].split(",")
                for otu in spikein2otus[spikein]
            ]
        else:
            otus_spikein = []
        if otus_spikein:
            spikein_reads = otu_count_series.loc[otus_spikein].sum()
            otu_count_series = otu_count_series.drop(otus_spikein)
        else:
            spikein_reads = -1

        norm_factor = spikein_reads if otus_spikein else 1
        if sample_weight_key and row[sample_weight_key]:
            norm_factor *= row[sample_weight_key]
        return pd.Series(
            {
                "spikein_reads": spikein_reads,
                "non_spikein_reads": otu_count_series.sum(),
                "norm_factor": norm_factor,
            }
        )

    if df_meta.shape[1]:
        df_meta_add = df_meta.apply(compute_spikein_and_norm, axis=1)
    else:
        df_meta_add = pd.DataFrame(
            {
                "spikein_reads": -1,
                "non_spikein_reads": df_otu_count.sum(axis=1),
                "norm_factor": 1,
            }
        )
    df_meta = df_meta.join(df_meta_add)
    return df_meta


# def _calc_alpha_metrics(df: pd.DataFrame) -> pd.DataFrame:
#     """Calculate alpha diversity metrics for each sample.
#     Note that you should ignore chao1 if your input data is not integer count.
#     """
#     return pd.DataFrame(
#         {
#             "chao1": df.apply(chao1, axis=1),
#             "richness": df.apply(lambda x: (x > 0).sum(), axis=1),
#             "shannon": df.apply(shannon, axis=1),
#             "simpson": df.apply(simpson, axis=1),
#         },
#         index=df.index,
#     )


def _agg_along_axis(
    df: pd.DataFrame, series: pd.Series, axis: int, aggfunc: str = None
) -> pd.DataFrame:
    """
    Aggregate a DataFrame along a specified axis using a given function.

    Parameters
    ----------
    df : pd.DataFrame
        The input DataFrame to be aggregated.
    series : pd.Series
        A Series used for grouping the DataFrame.
    axis : int
        The axis along which to perform the aggregation.
        0 for rows and 1 for columns.
    aggfunc : str, optional
        The aggregation function to use. Must be either 'mean' or 'sum'.
        If None, the function defaults to mean for axis 0 and sum for axis 1.

    Returns
    -------
    pd.DataFrame
        The aggregated DataFrame with groups based on the series provided.

    Raises
    ------
    ValueError
        If `aggfunc` is not 'mean' or 'sum'.
    """
    index_name = df.index.name or "index"
    series_index_name = series.index.name or "index"
    df = pl.from_pandas(df.reset_index().rename(columns={index_name: "row"}))
    exprs = [pl.col("rel_ab").mean(), pl.col("rel_ab").sum()]

    if not aggfunc:
        expr = exprs[axis]
    elif aggfunc == "sum":
        expr = exprs[1]
    elif aggfunc == "mean":
        expr = exprs[0]
    else:
        raise ValueError("expr must be either 'mean' or 'sum'.")

    if axis == 0:
        series = pl.from_pandas(
            series.reset_index(name="group").rename(columns={series_index_name: "row"})
        )
        df_g = (
            df.unpivot(index="row", variable_name="col", value_name="rel_ab")
            .join(series, on="row")
            .group_by(["col", "group"])
            .agg(rel_ab=expr)
            .pivot(
                index="group",
                on="col",
                values="rel_ab",
            )
            .to_pandas()
            .set_index("group")
        )
    else:
        series = pl.from_pandas(
            series.reset_index(name="group").rename(columns={series_index_name: "col"})
        )
        df_g = (
            df.unpivot(index="row", variable_name="col", value_name="rel_ab")
            .join(series, on="col")
            .group_by(["row", "group"])
            .agg(rel_ab=expr)
            .pivot(
                index="group",
                on="row",
                values="rel_ab",
            )
            .to_pandas()
            .set_index("group")
        ).T
    # on, groupby = ("row", "col") if axis == 0 else ("col", "row")
    # df_g = (
    #     df.melt(id_vars="row", variable_name="col", value_name="rel_ab")
    #     .join(series, on=on)
    #     .group_by([groupby, "group"])
    #     .agg(rel_ab=pl.col("rel_ab").mean())
    #     .pivot(
    #         index="group",
    #         columns=groupby,
    #         values="rel_ab",
    #     )
    #     .to_pandas()
    #     .set_index("group")
    # )
    # df_g = df_g if axis == 0 else df_g.T
    df_g.index.name = index_name
    return df_g


def _taxa_qc(
    df_tax_rel_ab: pd.DataFrame,
    rel_ab_thres: float,
    keep_rare: bool,
    keep_unknown: bool,
) -> pd.DataFrame:
    """
    Perform quality control on taxonomic data by filtering rare and unknown taxa.

    Parameters
    ----------
    df_tax_rel_ab : pd.DataFrame
        A DataFrame containing the relative abundance of taxa. Columns represent
            different taxa, and rows represent samples.
    rel_ab_thres : float
        Threshold to determine rare taxa. If float, taxa with relative abundance below
            this threshold in all samples are considered rare.
        If int, taxa with a rank of relative abundance above this threshold in all
            samples are considered rare.
    keep_rare : bool
        If False, remove rare taxa. If True, combine rare taxa into a single 'others' column.
    keep_unknown : bool
        If False, remove taxa labeled as unknown. If True, keep unknown taxa.

    Returns
    -------
    pd.DataFrame
        A DataFrame with filtered taxa based on the provided thresholds and options.

    Notes
    -----
    A taxon is defined as rare:
    - When `rel_ab_thres` is a float: If its relative abundance is below the threshold in all samples.
    - When `rel_ab_thres` is an int: If the rank of its relative abundance is above the threshold in all samples.

    A taxon is defined as unknown if its name contains 'unknown' or ends with '_UNKNOWN'.
    """
    df_tax_rel_ab_temp = df_tax_rel_ab[
        [
            t
            for t in df_tax_rel_ab.columns
            if not t.endswith("_UNKNOWN") and t != "unknown"
        ]
    ]
    if rel_ab_thres and int(rel_ab_thres) == rel_ab_thres:
        rare_taxa = df_tax_rel_ab_temp.columns[
            (
                df_tax_rel_ab_temp.rank(axis=1, ascending=False, method="max")
                > rel_ab_thres
            ).all(axis=0)
        ].tolist()
    else:
        rare_taxa = df_tax_rel_ab_temp.columns[
            (df_tax_rel_ab_temp < rel_ab_thres).all(axis=0)
        ].tolist()
    # if "unknown" in rare_taxa:
    #     rare_taxa.remove("unknown")
    # rare_taxa = [t for t in rare_taxa if not t.endswith("_UNKNOWN") or t == "unknown"]
    if not keep_rare:
        df_tax_rel_ab = df_tax_rel_ab.drop(rare_taxa, axis=1)
    elif rare_taxa:
        df_tax_rel_ab_rare = (
            df_tax_rel_ab[rare_taxa].sum(axis=1).to_frame(name="others")
        )
        df_tax_rel_ab = df_tax_rel_ab.drop(rare_taxa, axis=1)
        df_tax_rel_ab = pd.concat([df_tax_rel_ab, df_tax_rel_ab_rare], axis=1)
    if not keep_unknown:
        df_tax_rel_ab = df_tax_rel_ab[
            df_tax_rel_ab.columns[
                ~df_tax_rel_ab.columns.str.lower().str.endswith("unknown")
            ]
        ]
        if (df_tax_rel_ab.dtypes != "int").any():
            df_tax_rel_ab = df_tax_rel_ab.div(df_tax_rel_ab.sum(1), 0).fillna(0)
    return df_tax_rel_ab


def read_tables(
    otu_count_table: str,
    metadata_path: str = None,
    otu_taxonomy_path: str = None,
    warning: bool = True,
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    This function reads several tables from amplicon analysis, does necessary
        preprocessing, and return tables as dataframes.

    It reads OTU count table output by `VSEARCH` or `USERACH` where rows are sample
        names and columns are OTU names. The header of the first column is `#OTU ID`.

    Optionally, it takes a metadata table whose first column should be sample name. It
        ensures each sample in the metadata has associated OTU counts. If some samples
        are absent in the OTU count table, it fills in missing samples with all zero
        counts. Samples not in the metadata will be removed from the OTU count table.
        Samples in the returned OTU count table will be reordered according to the
        metadata.

    Optionally, it takes a taxonomy table whose the first column is OTU names and
        columns are taxonomy levels, and each taxonomy level has a paired confidence
        column. For example, the level `geuns` has a paired confidence column `geuns_p`.
    This function takes the intersection of OTUs in the OTU count table and the taxonomy
        table. If some OTUs do not have taxonomy, their count will be aggregated to a
        new "unknown" column, and a new row will be added to the taxonomy dataframe
        where taxonomy levels are "unknown" and confidences are -1.
    This function also copies index to a new column `otu` and `otu_p` if not already
        present.

    If sample metadata is not provided, metadata dataframe will be initialized from the
        samples in the OTU count table and have no columns.
    If taxonomy is not provided, taxonomy dataframe will be initialized from the OTUs in
        the OTU count table, and with only `otu` and `otu_p` columns.

    Returns:
        OTU count table: pd.DataFrame. Rows are samples, columns are OTUs, transposed
            from the input table.
        Sample metadata: pd.DataFrame.
        OTU taxonomy: pd.DataFrame.
    """
    df_otu_count = read_table(otu_count_table, index_col="#OTU ID")
    if not df_otu_count.columns.is_unique:
        raise ValueError("Sample names in OTU count table must be unique.")

    if metadata_path is not None:
        df_meta = read_table(metadata_path, index_col="sample", comment="#")
        if not df_meta.index.is_unique:
            raise ValueError("Sample names in metadata must be unique.")

        # add missing samples with all zero counts
        missing_samples = list(set(df_meta.index) - set(df_otu_count.columns))
        if missing_samples:
            if warning:
                print(
                    f"WARNING: The following {len(missing_samples)} samples are "
                    "missing from OTU count table:"
                )
                print("\t", end="")
                print(", ".join(map(str, sorted(missing_samples))))
                print("Filling in missing samples with all zero counts.")
            df_otu_count = pd.concat(
                [
                    df_otu_count,
                    pd.DataFrame(0, index=df_otu_count.index, columns=missing_samples),
                ],
                axis=1,
            )
        # only care about samples in metadata
        df_otu_count = df_otu_count[df_meta.index].transpose()
    else:
        df_otu_count = df_otu_count.transpose()
        df_meta = pd.DataFrame(index=df_otu_count.index)

    if otu_taxonomy_path is not None:
        # df_tax = pd.read_table(otu_taxonomy_path, index_col="otu").rename(
        #     {"otu.1": "otu"}, axis=1
        # )
        df_tax = read_table(otu_taxonomy_path, index_col="otu")
        if not df_tax.index.is_unique:
            raise ValueError("OTU numbers in taxonomy table must be unique.")
        # only care about OTUs with taxonomy
        otu_in_tax = set(df_tax.index)
        otu_in_count = set(df_otu_count.columns)
        no_tax_otus = list(otu_in_count - otu_in_tax)
        no_count_otus = list(otu_in_tax - otu_in_count)
        if no_tax_otus:
            common_otus = list(otu_in_tax & otu_in_count)
            df_tax = df_tax.loc[common_otus]
            df_tax_add = pd.DataFrame(
                "unknown",
                index=no_tax_otus,
                columns=df_tax.columns,
            )
            df_tax_add.loc[:, df_tax.columns[df_tax.columns.str.endswith("_p")]] = -1
            df_tax = pd.concat([df_tax, df_tax_add]).loc[df_otu_count.columns]
        if no_count_otus:  # add missing OTUs with all zero counts
            df_otu_count = pd.concat(
                [
                    df_otu_count,
                    pd.DataFrame(0, columns=no_count_otus, index=df_otu_count.index),
                ],
                axis=1,
            )
    else:
        df_tax = pd.DataFrame(index=df_otu_count.columns)
    if "otu" not in df_tax.columns:
        df_tax["otu"] = df_tax.index
    if "otu_p" not in df_tax.columns:
        df_tax["otu_p"] = 1
    return df_otu_count, df_meta, df_tax


def find_spikein_otus(
    df_meta: pd.DataFrame,
    df_tax: pd.DataFrame,
    spikein_taxa_key: str = None,
    only_otus: bool = False,
) -> dict[str, list[str]] | list[str]:
    if spikein_taxa_key is None:
        return []
    try:
        all_spikeins = (
            df_meta[spikein_taxa_key]
            .dropna()
            .str.split(",")
            .explode()
            .unique()
            .tolist()
        )
    except AttributeError:  # happens when all spikein_taxa_key are NaN
        all_spikeins = []
    spikein2otus = {
        spikein: _find_otus_by_taxon(df_tax, spikein) for spikein in all_spikeins
    }
    if only_otus:
        return list(set([otu for otus in spikein2otus.values() for otu in otus]))
    else:
        return spikein2otus


def get_otu_count(
    otu_count_table: str,
    metadata_path: str | None = None,
    otu_taxonomy_path: str | None = None,
    sample_weight_key: str | None = None,
    spikein_taxa_key: str | None = None,
    warning: bool = True,
) -> pd.DataFrame:
    df_otu_count, df_meta, df_tax = read_tables(
        otu_count_table, metadata_path, otu_taxonomy_path, warning=warning
    )
    df_meta = _calc_norm_factor(
        df_otu_count,
        df_meta,
        df_tax,
        sample_weight_key,
        spikein_taxa_key,
    )
    df_meta["sequencing_depth"] = df_otu_count.sum(
        axis=1
    ) - df_meta.spikein_reads.to_numpy().clip(0)
    df_otu_count = df_otu_count.drop(
        find_spikein_otus(df_meta, df_tax, spikein_taxa_key, only_otus=True), axis=1
    )
    return df_otu_count, df_meta, df_tax


def main():
# if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    # Arguments
    parser.add_argument(
        "-i", "--otu_count_tsv_path", required=True, help="Path to OTU count tsv."
    )
    parser.add_argument(
        "-m", "--metadata_path", required=True, help="Path to sample metadata."
    )
    parser.add_argument(
        "-t", "--otu_taxonomy_path", required=True, help="Path to OTU taxonomy tsv."
    )
    parser.add_argument("-sp", "--spikein_taxa_key", type=str, default="spike_in")
    parser.add_argument("-r", "--rep_group_key", type=str, default="rep_group")
    parser.add_argument("-w", "--sample_weight_key", default=None, type=str)
    parser.add_argument(
        "-o", "--output_dir", required=True, help="Directory to save the outputs."
    )
    parser.add_argument(
        "-l",
        "--tax_levels",
        nargs="+",
        required=True,
        help="Taxonomy level of interest, must be a column in the taxonomy table.",
    )
    parser.add_argument(
        "-a",
        "--rel_ab_thresholds",
        nargs="+",
        type=float,
        help="Genus relative abundance threshold.",
    )
    parser.add_argument("--keep_others", action="store_true", default=False)
    parser.add_argument("--keep_unknown", action="store_true", default=False)
    args = parser.parse_args()

    # Extract arg values
    count_tsv_path = args.otu_count_tsv_path
    metadata_path = args.metadata_path
    otu_taxonomy_path = args.otu_taxonomy_path
    spikein_taxa_key = args.spikein_taxa_key
    sample_weight_key = args.sample_weight_key
    output_dir = args.output_dir
    rep_group_key = args.rep_group_key
    tax_levels = args.tax_levels
    rel_ab_thresholds = args.rel_ab_thresholds
    keep_others = args.keep_others
    keep_unknown = args.keep_unknown

    os.makedirs(output_dir, exist_ok=True)

    # read inputs
    df_otu_count, df_meta, df_tax = get_otu_count(
        count_tsv_path,
        metadata_path,
        otu_taxonomy_path,
        sample_weight_key,
        spikein_taxa_key,
    )

    df_meta.to_csv(f"{output_dir}/metadata_sample.csv")
    df_otu_abs_ab = df_otu_count.div(df_meta.norm_factor, axis=0)
    df_otu_abs_ab.to_csv(f"{output_dir}/ab_abs_s_otu.csv")

    df_otu_rel_ab = df_otu_count.div(df_otu_count.sum(axis=1), axis=0)
    df_otu_rel_ab_g = _agg_along_axis(df_otu_rel_ab, df_meta[rep_group_key], axis=0)

    # df_otu_count.to_csv(f"{output_dir}/count_sample_otu.csv")
    # absolute abundance is only for OTU level

    if len(rel_ab_thresholds) == 1:
        rel_ab_thresholds = rel_ab_thresholds * len(tax_levels)
    for level, rel_ab_thres in zip(tax_levels, rel_ab_thresholds):
        df_tax_rel_ab_g = _taxa_qc(
            _agg_along_axis(df_otu_rel_ab_g, df_tax[level], axis=1),
            rel_ab_thres,
            keep_others,
            keep_unknown,
        )
        df_tax_rel_ab = _taxa_qc(
            _agg_along_axis(df_otu_rel_ab, df_tax[level], axis=1),
            rel_ab_thres,
            keep_others,
            keep_unknown,
        )
        df_tax_rel_ab_g.to_csv(f"{output_dir}/ab_rel_g_{level}.csv")
        df_tax_rel_ab.to_csv(f"{output_dir}/ab_rel_s_{level}.csv")
